//**********************************************************************
// This source code is provided as an example.  No warranty is expressed
// or implied.  It is your responsibility to determine its suitability
// to your application, to modify it as necessary, and to validate
// its correctness.
//
// Licensed Materials - Property of IBM
// (c) Copyright IBM Corporation 2000. All Rights Reserved.
//
// US Government Users Restricted Rights - Use, duplication, or
// disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
//**********************************************************************
// Revision / History
// Defect     Date       Who      Description
// ---------- 01/03/2000 ldbader  Module created
//**********************************************************************

package com.ibm.etill.ldbcardcassette;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import com.ibm.etill.framework.cassette.AcceptPaymentRequest;
import com.ibm.etill.framework.cassette.AdminRequest;
import com.ibm.etill.framework.cassette.APIRequest;
import com.ibm.etill.framework.cassette.APIResponse;
import com.ibm.etill.framework.cassette.ApproveRequest;
import com.ibm.etill.framework.cassette.ApproveReversalRequest;
import com.ibm.etill.framework.cassette.BatchRequest;
import com.ibm.etill.framework.cassette.BatchCloseRequest;
import com.ibm.etill.framework.cassette.BatchOpenRequest;
import com.ibm.etill.framework.cassette.BatchPurgeRequest;
import com.ibm.etill.framework.cassette.CancelOrderRequest;
import com.ibm.etill.framework.cassette.CloseOrderRequest;
import com.ibm.etill.framework.cassette.Cassette;
import com.ibm.etill.framework.cassette.CassetteBatch;
import com.ibm.etill.framework.cassette.CassetteOrder;
import com.ibm.etill.framework.cassette.CassetteRequest;
import com.ibm.etill.framework.cassette.CassetteResponse;
import com.ibm.etill.framework.cassette.CassetteTransaction;
import com.ibm.etill.framework.cassette.DeleteBatchRequest;
import com.ibm.etill.framework.cassette.DepositRequest;
import com.ibm.etill.framework.cassette.DepositReversalRequest;
import com.ibm.etill.framework.cassette.ProtocolRequest;
import com.ibm.etill.framework.cassette.OrderRequest;
import com.ibm.etill.framework.cassette.ReceivePaymentRequest;
import com.ibm.etill.framework.cassette.RefundRequest;
import com.ibm.etill.framework.cassette.RefundReversalRequest;
import com.ibm.etill.framework.io.ETillConnection;
import com.ibm.etill.framework.log.ErrorLog;
import com.ibm.etill.framework.log.ETillCassetteException;
import com.ibm.etill.framework.log.Trace;
import com.ibm.etill.framework.payapi.AliasValidationItem;
import com.ibm.etill.framework.payapi.ETillAbortOperation;
import com.ibm.etill.framework.payapi.Batch;
import com.ibm.etill.framework.payapi.Credit;
import com.ibm.etill.framework.payapi.IntegerValidationItem;
import com.ibm.etill.framework.payapi.Order;
import com.ibm.etill.framework.payapi.Payment;
import com.ibm.etill.framework.payapi.ParameterTable;
import com.ibm.etill.framework.payapi.StringValidationItem;
import com.ibm.etill.framework.supervisor.Supervisor;

/***********************************************************************
*  The Cassette object is responsible for managing the cassettes
*  resources within the framework.  The framework
*  uses Cassette methods for initialization, creating new objects,
*  resurrecting existing objects from the database, and for servicing
*  merchant requests.
* <p>
* In many cases the Cassette object uses other objects to perform
* its responsiblities.  For example, AcceptPayment requests are sent
* to the cassette's Order object for processing.
***********************************************************************/
public final class LdbCardCassette
   extends Cassette
   implements LdbCardConstants
{

   private static final String COPYRIGHT =
      "(c) Copyright IBM Corporation 2000 - All Rights Reserved.";

   //<gutter>
   //</gutter>

   //*******************************************************************
   //This is the cassette's class data.
   //******************************************************************/
   /********************************************************************
   * This cassette class attribute contains the hashtable that defines
   * the validation parameters for cassette dependent protocol data.
   * This hashtable is initialized in a static block.
   ********************************************************************/
   private static final Hashtable clsValidationTable = new Hashtable();

// /********************************************************************
// * This static block initializes the validation table.  It is run
// * once and only once when the class is first loaded into the
// * Java virtual machine.
// ********************************************************************/
   static {

      clsValidationTable.put(
         PD_PAY_PAN,
         new LuhnStringValidationItem(
            PD_PAY_PAN,                     // $PAN
            RC_CASSETTE_PAN,                // 1015
            false,                          // No null values
            12,                             // Minimum length
            19                              // Maximum length
         )
      );
      clsValidationTable.put(
         PD_PAY_EXPIRY,
         new StringValidationItem(
            PD_PAY_EXPIRY,                  // $EXPIRY
            RC_CASSETTE_EXPIRY,             // 1016
            false,                          // No null values
            6,                              // Minimum length
            6,                              // Maximum length
            "8859_1"                        // Standard 8 bit ASCII
         )
      );
      clsValidationTable.put(
         PD_PAY_BRAND,
         new StringValidationItem(
            PD_PAY_BRAND,                   // $BRAND
            RC_CASSETTE_BRAND,              // 1014
            false,                          // No null values
            2,                              // Minimum length
            40,                             // Maximum length
            "8859_1"                        // Standard 8 bit ASCII
         )
      );
      clsValidationTable.put(
         PD_PAY_CARDHOLDERNAME,
         new StringValidationItem(
            PD_PAY_CARDHOLDERNAME,          // $CARDHOLDERNAME
            SRC_CASSETTE_CARDHOLDERNAME,    //
            false,                          // No null values
            2,                              // Minimum length
            80,                             // Maximum length
            "UTF8"                          // UCS Transformation Format
         )
      );
      clsValidationTable.put(
         PD_PAY_STREETADDRESS,
         new StringValidationItem(
            PD_PAY_STREETADDRESS,           // $AVS.STREETADDRESS
            RC_CASSETTE_AVS_STREETADDRESS,  // 1053
            false,                          // No null values
            2,                              // Minimum length
            80,                             // Maximum length
            "UTF8"                          // UCS Transformation Format
         )
      );
      clsValidationTable.put(
         PD_PAY_STREETADDR,
         new AliasValidationItem(
            PD_PAY_STREETADDR,              // $AVS.STREETADDR for
            PD_PAY_STREETADDRESS            // Version 1.2 compatibility
         )
      );
      clsValidationTable.put(
         PD_PAY_CITY,
         new StringValidationItem(
            PD_PAY_CITY,                    // $AVS.CITY
            RC_CASSETTE_AVS_CITY,           // 1054
            false,                          // No null values
            2,                              // Minimum length
            80,                             // Maximum length
            "UTF8"                          // UCS Transformation Format
         )
      );
      clsValidationTable.put(
         PD_PAY_STATEPROVINCE,
         new StringValidationItem(
            PD_PAY_STATEPROVINCE,           // $AVS.STATEPROVINCE
            RC_CASSETTE_AVS_STATEPROVINCE,  // 1055
            false,                          // No null values
            2,                              // Minimum length
            80,                             // Maximum length
            "UTF8"                          // UCS Transformation Format
         )
      );
      clsValidationTable.put(
         PD_PAY_STATEPROV,
         new AliasValidationItem(
            PD_PAY_STATEPROV,               // $AVS.STATEPROV for
            PD_PAY_STATEPROVINCE            // Version 1.2 compatibility
         )
      );
      clsValidationTable.put(
         PD_PAY_COUNTRYCODE,
         new StringValidationItem(
            PD_PAY_COUNTRYCODE,             // $AVS.COUNTRYCODE
            RC_CASSETTE_AVS_COUNTRYCODE,    // 1052
            false,                          // No null values
            2,                              // Minimum length
            80,                             // Maximum length
            "UTF8"                          // UCS Transformation Format
         )
      );
      clsValidationTable.put(
         PD_PAY_POSTALCODE,
         new StringValidationItem(
            PD_PAY_POSTALCODE,              // $AVS.POSTALCODE
            RC_CASSETTE_AVS_POSTALCODE,     // 1056
            false,                          // No null values
            2,                              // Minimum length
            80,                             // Maximum length
            "UTF8"                          // UCS Transformation Format
         )
      );
      // The term AVS (Address Verification System) is unique to
      // North American VISA and Mastercard credit card networks.
      // Merchants may prefer to specify cardholder billing addresses
      // using the $BILL prefix because it is more universal:  it
      // applies to other geographies and to other payment instruments.
      // To improve cassette interchangability and simplify merchant
      // coding conventions, the $BILL protocol data should be defined
      // as an alias for $AVS protocol data on any cassette that
      // supports AVS.
      clsValidationTable.put(
         PD_PAY_B_STREETADDRESS,
         new AliasValidationItem(
            PD_PAY_B_STREETADDRESS,         // $BILL.STREETADDRESS =
            PD_PAY_STREETADDRESS            // $AVS.STREETADDRESS
         )
      );
      clsValidationTable.put(
         PD_PAY_B_STREETADDR,
         new AliasValidationItem(
            PD_PAY_B_STREETADDR,            // $BILL.STREETADDR for
            PD_PAY_STREETADDRESS            // Version 1.2 compatibility
         )
      );
      clsValidationTable.put(
         PD_PAY_B_CITY,
         new AliasValidationItem(
            PD_PAY_B_CITY,                  // $BILL.CITY =
            PD_PAY_CITY                     // $AVS.CITY
         )
      );
      clsValidationTable.put(
         PD_PAY_B_STATEPROVINCE,
         new AliasValidationItem(
            PD_PAY_B_STATEPROVINCE,         // $BILL.STATEPROVINCE =
            PD_PAY_STATEPROVINCE            // $AVS.STATEPRIVINCE
         )
      );
      clsValidationTable.put(
         PD_PAY_B_STATEPROV,
         new AliasValidationItem(
            PD_PAY_B_STATEPROV,             // $BILL.STATEPROV for
            PD_PAY_STATEPROVINCE            // Version 1.2 compatibility
         )
      );
      clsValidationTable.put(
         PD_PAY_B_COUNTRYCODE,
         new AliasValidationItem(
            PD_PAY_B_COUNTRYCODE,           // $BILL.COUNTRYCODE =
            PD_PAY_COUNTRYCODE              // $AVS.COUNTRYCODE
         )
      );
      clsValidationTable.put(
         PD_PAY_B_COUNTRYCOD,
         new AliasValidationItem(
            PD_PAY_B_COUNTRYCOD,            // $BILL.COUNTRYCOD for
            PD_PAY_COUNTRYCODE              // Version 1.2 compatibility
         )
      );
      clsValidationTable.put(
         PD_PAY_B_POSTALCODE,
         new AliasValidationItem(
            PD_PAY_B_POSTALCODE,            // $BILL.POSTALCODE =
            PD_PAY_POSTALCODE               // $AVS.POSTALCODE
         )
      );
      clsValidationTable.put(
         PD_ACC_CURRENCY,
         new StringValidationItem(
            PD_ACC_CURRENCY,                // $CURRENCY
            SRC_CASSETTE_CURRENCY,          //
            false,                          // No null values
            3,                              // Minimum length
            3,                              // Maximum length
            "UTF8"                          // UCS Transformation Format
         )
      );
      clsValidationTable.put(
         PD_ACC_USER,
         new StringValidationItem(
            PD_ACC_USER,                    // $USER
            SRC_CASSETTE_USER,              //
            false,                          // No null values
            4,                              // Minimum length
            10,                             // Maximum length
            "UTF8"                          // UCS Transformation Format
         )
      );
      clsValidationTable.put(
         PD_ACC_PASSWORD,
         new StringValidationItem(
            PD_ACC_PASSWORD,                // $PASSWORD
            SRC_CASSETTE_PASSWORD,          //
            false,                          // No null values
            4,                              // Minimum length
            10,                             // Maximum length
            "UTF8"                          // UCS Transformation Format
         )
      );
      clsValidationTable.put(
         PD_ACC_LDBCARDID,
         new StringValidationItem(
            PD_ACC_LDBCARDID,               // $LDBCARDID
            SRC_CASSETTE_LDBCARDID,         //
            false,                          // No null values
            4,                              // Minimum length
            10,                             // Maximum length
            "UTF8"                          // UCS Transformation Format
         )
      );
      clsValidationTable.put(
         PD_ACC_URL,
         new StringValidationItem(
            PD_ACC_URL,                     // $LDBCARDRMIURL
            SRC_CASSETTE_LDBCARDRMIURL,     //
            false,                          // No null values
            4,                              // Minimum length
            200,                            // Maximum length
            "UTF8"                          // UCS Transformation Format
         )
      );
      clsValidationTable.put(
         PD_ACC_BATCH_CLOSE_TIME,
         new IntegerValidationItem(
            PD_ACC_BATCH_CLOSE_TIME,        // $BATCHCLOSETIME
            SRC_CASSETTE_BATCH_CLOSE_TIME,  //
            true,                           // Null values allowed
            new Integer(0),                 // 1st second of the day
            new Integer(86399)              // Last second of the day
         )
      );
      clsValidationTable.put(
         PD_ACC_MIN_BATCH_TOTAL,
         new IntegerValidationItem(
            PD_ACC_MIN_BATCH_TOTAL,         // $MINBATCHTOTAL
            SRC_CASSETTE_MIN_BATCH_TOTAL,   //
            true,                           // Null values allowed
            new Integer(0),                 // smallest minimum
            null                            // do not validate maximum
         )
      );
      clsValidationTable.put(
         PD_ACC_MAX_BATCH_SIZE,
         new IntegerValidationItem(
            PD_ACC_MAX_BATCH_SIZE,          // $MAXBATCHSIZE
            SRC_CASSETTE_MAX_BATCH_SIZE,    //
            true,                           // Null values allowed
            new Integer(0),                 // smallest minimum
            null                            // do not validate maximum
         )
      );

   }

   /***********************************************************************
   * This class converts a parameter's value from a byte array into a
   * string and then validates that the result passes the Luhn check.
   * <p>
   * Credit Card numbers are (most times) 13 to 16 digit numbers which are
   * protected by a special numerical check, called Luhn check.  From right
   * to left, each digit is multiplied by the alternating factors 1 and 2
   * (the rightmost digit is always multiplied by 1).  For each result, the
   * digits are summed together.  Then, these sums are totaled.
   * Finally, to be a valid Credit Card number, the total must be evenly
   * divisible by 10.
   * <p>
   *   Example (Credit Card number: 1234 5678 7654 3210):
   * <p>
   *   First, start from the right and multiply each digit by an alternating
   *   factor of 1 and 2.
   *   1   2   3   4   5   6   7   8   7   6   5   4   3   2   1   0
   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *
   *   2   1   2   1   2   1   2   1   2   1   2   1   2   1   2   1
   *   =   =   =   =   =   =   =   =   =   =   =   =   =   =   =   =
   *   2   2   6   4  10  12  14  16  14  12  10   4   6   2   2   0
   * <p>
   *   Sum the digits in the results.
   *   2   2   6  4 (1+0) (1+2) (1+4) (1+6) (1+4) (1+2) (1+0) 4  6  2  2  0
   * <p>
   *   Total the sums from above.
   *   2 + 2 + 6 + 5 + 1 + 3 + 5 + 7 + 5 + 3 + 1 + 4 + 6 + 2 + 2 + 0 = 60
   * <p>
   *   Divide the sum by 10 and if the remainder is 0 the value is valid.
   *   60 % 10 = 0 so this number is valid.
   ***********************************************************************/
   static public class LuhnStringValidationItem
      extends StringValidationItem
   {

      /********************************************************************
      * The Luhn value of a given digit depends on the alternating factor
      * that is applied to the digit.  The top half (first 10 values)
      * in this array provides the Luhn value of the respective digit when
      * the alternating factor is 1.  The bottom half (last 10 values)
      * in this array provides the Luhn value of the respective digit when
      * the alternating factor is 2.
      ********************************************************************/
      private static final int[] clsLuhnValue = {
         0,
         1,
         2,
         3,
         4,
         5,
         6,
         7,
         8,
         9,
         0,
         2,
         4,
         6,
         8,
         1,
         3,
         5,
         7,
         9,
      };

      /********************************************************************
      * This constructor is used to create a Luhn String Validation item.
      *
      * @param     argParameter - the parameter keyword whose value is to
      * be validated
      * @param     argSecondaryReturnCode - the secondary return code used
      * to identify the parameter when it is invalid
      * @param     argNullValuesAllowed - a flag that specifies whether or
      * not the parameter value can be null
      * @param     argMinLength - the minimum length of a valid string
      * @param     argMaxLength - the maximum length of a valid string
      ********************************************************************/
      public LuhnStringValidationItem(
         String         argParameter,
         short          argSecondaryReturnCode,
         boolean        argNullValuesAllowed,
         int            argMinLength,
         int            argMaxLength
      ) {
         // Initialize the parent class.  Note that the ISO 8859_1 encoding
         // can be safely used because only digits are allowed and the
         // validateAndInsertValue() method will make sure that only
         // digits were specified.
         super(
            argParameter,
            argSecondaryReturnCode,
            argNullValuesAllowed,
            argMinLength,
            argMaxLength,
            "8859_1"
         );
      }

      /********************************************************************
      * This method uses the parent's method to convert the byte array
      * into a string and validate that the result is not too long or
      * too short.  Note that the parent's method will place the result
      * into the parameter table if it does not detect an error.  Then
      * this method will continue the validation.  It will make sure that
      * each character is a digit and that it passes the Luhn test.
      *
      * @param     argValue - the value passed by the merchant
      * @param     argResultTable - the parameter table to be passed to the
      * cassette
      * @return  - a valid numeric string
      * @exception ETillAbortOperation - any error results in this
      * exception
      ********************************************************************/
      public Object validateAndInsertValue(
         byte[]         argValue,
         ParameterTable argResultTable
      ) throws ETillAbortOperation {

         // Use the parent class to convert the value into a string
         // and verify the length.
         String varResult = (String)super.validateAndInsertValue(
            argValue,
            argResultTable
         );

         // Continue verification if the result is not null.
         if (varResult != null) {
            // Inititalize the Luhn value.
            int varAccumulatedLuhnValue = 0;
            // Walk through the string from right to left, verify that
            // each character is a digit, and accumulate the Luhn value.
            for (
               int varIndex = varResult.length() - 1,
               varAlternatingOffset = 0;
               varIndex >= 0;
               varIndex--,
               varAlternatingOffset = (varAlternatingOffset == 0) ? 10 : 0
            ) {
               char varChar = varResult.charAt(varIndex);
               if (
                  varChar < '0' ||
                  varChar > '9'
               ) {
                  throw new ETillAbortOperation(
                     PRC_PARAMETER_FORMAT_ERROR,
                     this.returnCode
                  );
               }
               varAccumulatedLuhnValue +=
                  clsLuhnValue[ (varChar - '0') + varAlternatingOffset ];
            }

            // If every digit was 0 or if the remainder after dividing by
            // 10 is not zero, the value is invalid.
            if (
               varAccumulatedLuhnValue == 0 ||
               (varAccumulatedLuhnValue % 10) != 0
            ) {
               throw new ETillAbortOperation(
                  PRC_PARAMETER_VALUE_ERROR,
                  this.returnCode
               );
            }
         }

         return varResult;
      }

   }

   /********************************************************************
   * This constructor is used by the framework during initialization.
   * Since the class is loaded dynamically, the no-argument constructor
   * is always used.
   *<p>
   * NOTE:  The framework has not initialized the cassette when this
   * constructor is called.  This method cannot issue trace, log, or
   * database commands at this point. All cassette initialization should
   * be performed in subsequent initialization calls.
   ********************************************************************/
   public LdbCardCassette() {
   }

   /********************************************************************
   * This method is used by the framework to give the cassette its
   * resource bundle identifier.  Although this method is maintained
   * for compatibilty reasons, it is no longer required.  Payment
   * Manager 2.1 now uses a string to reference the resource bundle
   * and the string is always set to the cassette name.  Since this
   * cassette assumes that its classes are all final, the cassette
   * name is a constant, so it simply makes an assertion test to make
   * sure that the resource bundle loaded by the framework is the same
   * as the resource bundle constant used in the cassette.
   * <p>
   * The resource bundle for the cassette is automatically loaded by
   * the framework before any other initialization is attempted.  The
   * name of the resource bundle is derived from the payment type
   * defined for the cassette.  For example, if you are implementing
   * a cassette for payment type Xxxx, you must supply a default
   * property file of the form Xxxx.properties.  Additional property
   * files can be provided for national languages other than your
   * default language whenever required, without affecting any Java
   * code in your package.  See the JavaDoc for java.util.ResourceBundle
   * for more information.
   *
   * @param     argResourceBundleId - an integer, assigned by the
   * framework, that represents the resource bundle that has been
   * loaded for this cassette.
   ********************************************************************/
   public void setMessageID(
      String argId
   ) {

      // Assertion test:  If the resource bundle loaded by the framework
      // is different from the contant used to identify the resource
      // bundle in this cassette, there is a serious defect and the
      // cassette cannot continue.
      if ( !(RB_ID.equals(argId)) ) {
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardCassette Resource bundle ID from framework does" +
               " not match the constant used in the cassette" +
               "\n\tPRC = " + PRC_CASSETTE_ERROR +
               "\n\tSRC = " + SRC_CASSETTE_BUNDLE_ID_MISMATCH
            );
         }
         ErrorLog.logError(
            RB_ID,
            MSG_CASSETTE_BUNDLE_ID_MISMATCH,
            null,
            RB_ID,
            argId
         );
         throw new ETillCassetteException();
      }
   }


   /********************************************************************
   * This method is used by the framework during initialization.  The
   * cassette must read its configuration and throw an exception if
   * a failure occurs.
   * <p>
   * For this cassette, the account class reads the configuration and
   * constructs an account object for each row in the account
   * configuration table.
   *
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public void readCassetteConfig()
      throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardCassette.readCassetteConfig()"
         );
      }

      LdbCardAccount.resurrectAccounts();

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardCassette.readCassetteConfig()"
         );
      }
   }

   /********************************************************************
   * This method is used by the framework during initialization.  It is
   * called after a successful call to readCassetteConfig().  The
   * cassette must perform any verification that cannot be done
   * when the configuration was read.  For example, this method would
   * verify that related configuration objects are consistent.
   * <p>
   * For this cassette, the cassette uses its account class to verify
   * that the configuration is correct.
   *
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public void verifyConfig()
      throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardCassette.verifyConfig()"
         );
      }

      LdbCardAccount.verifyAccounts();

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardCassette.verifyConfig()"
         );
      }
   }

   /********************************************************************
   * This method is used by the framework during initialization.  It is
   * called after a successful call to verifyConfig().  The cassette
   * must complete the remaining initialization for any objects
   * created from the configuration.
   * <p>
   * For this cassette, the cassette has each account object initialize.
   *
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public void initializeCassette()
      throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardCassette.initializeCassette()"
         );
      }

      LdbCardAccount.initializeAccounts();

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardCassette.initializeCassette()"
         );
      }
   }

   /********************************************************************
   * This method is used by the framework during initialization.  It is
   * called after a successful call to initializeCassette().  The
   * cassette must return a vector of ComPoint objects used to service
   * I/O connections (TCP/IP connections to the payment engine, for
   * example).  ComPoint objects are required for asynchronous
   * communication.  The frame work builds a thread pool for each
   * ComPoint object.  When data is received, a thread is assigned to
   * process the data flow.
   * <p>
   * For this cassette, the data exchange between the cassette and the
   * payment engine is synchronous.  The thread processing the merchant
   * request is held while waiting for a response from the payment
   * engine. ComPoint objects are <b>not</b> used, so an empty Vector
   * is returned.
   *
   * @return - A vector of ComPoint objects used to service cassette
   * I/O connections (TCP/IP connections to the payment engine, for
   * example).
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public Vector createComPoints()
      throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardCassette.createComPoints()"
         );
      }

      Vector varVector = new Vector();

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardCassette.createComPoints()"
         );
      }
      return varVector;
   }

   /********************************************************************
   * This method is used by the framework during initialization.  It is
   * called after a successful call to createComPoints().  This is the
   * cassette's last opportunity to perform initialization before
   * merchant requests will be processed.  Typically cassettes use
   * this opportunity to recover or restart any operation that was
   * in progress when the framework shut down.
   * <p>
   * For this cassette, all the payment operations are synchronous.  If
   * the framework is shutdown in the middle of a transaction,
   * it is as if the transaction was never performed.  The merchant
   * never receives confirmation and retries the operation.  Therefore
   * this method merely logs a message indicating that initializtion
   * is complete.
   ********************************************************************/
   public void start() {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardCassette.start()"
         );
      }

      ErrorLog.logError(
         RB_ID,
         MSG_CASSETTE_HAS_STARTED,
         null
      );


      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardCassette.start()"
         );
      }
   }

   /********************************************************************
   * This method is used by the framework to obtain a hashtable that
   * contains an element for each cassette dependent protocol data
   * keyword that is to be validated and converted into the proper]
   * data type by the framework before it is passed to the cassette
   * in a parameter table.  Each element in the table has a key
   * that contains the protocol data keyword as a String and a value
   * that contains some subclass of the frameworks
   * ParameterValidationItem class.
   *
   * @return    A hashtable that contains an element for every
   * cassette dependent protocol data keyword to be validated and
   * converted automatically by the framework.
   ********************************************************************/
   public Hashtable getParameterValidationTable() {
      return clsValidationTable;
   }

   /********************************************************************
   * This method is called by the framework when an ReceivePayment or
   * AcceptPayment request is received.  The cassette must construct a
   * new cassette order object and return it to the framework.
   * <p>
   * For this cassette, a cassette order object is constructed and
   * returned.
   *
   * @param     argFrameworkOrder - a framework order object constructed
   * by the framework for a ReceivePayment or AcceptPayment request.
   * @param     argProtocolData - a hashtable containing the protocol
   * data values specified by the merchant.
   * @return    - a new cassette order object.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public CassetteOrder newCassetteOrder(
      Order                  argFrameworkOrder,
      ParameterTable         argProtocolData
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardCassette.newCassetteOrder()"
         );
      }

      LdbCardOrder varCassetteOrder = new LdbCardOrder(
         this,
         argFrameworkOrder,
         argProtocolData
      );

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardCassette.newCassetteOrder()"
         );
      }
      return varCassetteOrder;
   }

   /********************************************************************
   * This method is called by the framework when a request is received
   * for an order that has not been found in the cache. The cassette
   * must resurrect its matching order object from the information
   * previously recorded in the cassette order database table and
   * return the result to the framework.
   * <p>
   * For this cassette, a new cassette order is constructed using the
   * constructor that resurrects the order.  This method reads the
   * correct row from the cassette's order table and initializes
   * the object with the values found there.
   *
   * @param     argFrameworkOrder - a framework order object that has
   * been resurrected from the ETORDER database table.
   * @return    - a cassette order object that has been initialized
   * from the cassette order database table.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public CassetteOrder resurrectCassetteOrder(
      Order                  argFrameworkOrder
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardCassette.resurrectCassetteOrder()"
         );
      }

      LdbCardOrder varCassetteOrder = new LdbCardOrder(
         this,
         argFrameworkOrder
      );

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardCassette.resurrectCassetteOrder()"
         );
      }
      return varCassetteOrder;
   }

   /********************************************************************
   * This method is called by the framework when an Approve request is
   * received.  The cassette must construct a new cassette
   * payment object and return it to the framework.
   * <p>
   * For this cassette, a cassette payment object is constructed and
   * returned.
   *
   * @param     argFrameworkPayment - a framework payment object
   * constructed by the framework for an Approve request.
   * @param     argProtocolData - a hashtable containing the protocol
   * data values specified by the merchant.
   * @return    - a new cassette payment object.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public CassetteTransaction newCassettePayment(
      Payment                argFrameworkPayment,
      ParameterTable         argProtocolData
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardCassette.newCassettePayment()"
         );
      }

      LdbCardPayment varCassettePayment = new LdbCardPayment(
         argFrameworkPayment,
         argProtocolData
      );

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardCassette.newCassettePayment()"
         );
      }
      return varCassettePayment;
   }

   /********************************************************************
   * This method is called by the framework when a request is received
   * for a payment that has not been found in the cache. The cassette
   * must resurrect its matching payment object from the information
   * previously recorded in the cassette payment database table and
   * return the result to the framework.
   * <p>
   * For this cassette, a new cassette payment is constructed using the
   * constructor that resurrects the payment.  This method reads the
   * correct row from the cassette's payment table and initializes
   * the object with the values found there.
   *
   * @param     argFrameworkPayment - a framework payment object that
   * has been resurrected from the ETPAYMENT database table.
   * @return    - a cassette payment object that has been initialized
   * from the cassette payment database table.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public CassetteTransaction resurrectCassettePayment(
      Payment                argFrameworkPayment
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardCassette.resurrectCassettePayment()"
         );
      }

      LdbCardPayment varCassettePayment = new LdbCardPayment(
         argFrameworkPayment
      );

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardCassette.resurrectCassettePayment()"
         );
      }
      return varCassettePayment;
   }

   /********************************************************************
   * This method is called by the framework when a Refund request is
   * received.  The cassette must construct a new cassette
   * credit object and return it to the framework.
   * <p>
   * For this cassette, a cassette credit object is constructed and
   * returned.
   *
   * @param     argFrameworkCredit - a framework credit object
   * constructed the framework for a new Refund request.
   * @param     argProtocolData - a hashtable containing the protocol
   * data values specified by the merchant.
   * @return    - a new cassette credit object.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public CassetteTransaction newCassetteCredit(
      Credit                 argFrameworkCredit,
      ParameterTable         argProtocolData
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardCassette.newCassetteCredit()"
         );
      }

      LdbCardCredit varCassetteCredit = new LdbCardCredit(
         argFrameworkCredit,
         argProtocolData
      );

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardCassette.newCassetteCredit()"
         );
      }
      return varCassetteCredit;
   }

   /********************************************************************
   * This method is called by the framework when a request is received
   * for a credit that has not been found in the cache. The cassette
   * must resurrect its matching credit object from the information
   * previously recorded in the cassette credit database table and
   * return the result to the framework.
   * <p>
   * For this cassette, a new cassette credit is constructed using the
   * constructor that resurrects the credit.  This method reads the
   * correct row from the cassette's credit table and initializes
   * the object with the values found there.
   *
   * @param     argFrameworkCredit - a framework credit object that
   * has been resurrected from the ETCREDIT database table.
   * @return    - a cassette credit object that has been initialized
   * from the cassette credit database table.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public CassetteTransaction resurrectCassetteCredit(
      Credit                 argFrameworkCredit
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardCassette.resurrectCassetteCredit()"
         );
      }

      LdbCardCredit varCassetteCredit = new LdbCardCredit(
         argFrameworkCredit
      );

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardCassette.resurrectCassetteCredit()"
         );
      }
      return varCassetteCredit;
   }

   /********************************************************************
   * This method is called by the framework when an OpenBatch request
   * is received.  The cassette must contruct a new cassette batch
   * object and return it to the framework.
   * <p>
   * For this cassette, a cassette batch object is constructed and
   * returned.
   *
   * @param     argFrameworkBatch - a framework batch object
   * constructed by the framework for a BatchOpen request.
   * @param     argProtocolData - a hashtable containing the protocol
   * data values specified by the merchant.
   * @return    - a new cassette batch object.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public CassetteBatch newCassetteBatch(
      Batch                  argFrameworkBatch,
      ParameterTable         argProtocolData
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardCassette.newCassetteBatch()"
         );
      }

      LdbCardBatch varCassetteBatch = new LdbCardBatch(
         this,
         argFrameworkBatch,
         argProtocolData
      );

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardCassette.newCassetteBatch()"
         );
      }
      return varCassetteBatch;
   }

   /********************************************************************
   * This method is called by the framework when a request is received
   * for a batch that has not been found in the cache. The cassette
   * must resurrect its matching batch object from the information
   * previously recorded in the cassette batch database table and
   * return the result to the framework.
   * <p>
   * For this cassette, a new cassette batch is constructed using the
   * constructor that resurrects the batch.  This method reads the
   * correct row from the cassette's batch table and initializes
   * the object with the values found there.
   *
   * @param     argFrameworkBatch - a framework batch object that
   * has been resurrected from the ETBATCH database table.
   * @return    - a cassette batch object that has been initialized
   * from the cassette batch database table.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public CassetteBatch resurrectCassetteBatch(
      Batch                  argFrameworkBatch
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardCassette.resurrectCassetteBatch()"
         );
      }

      LdbCardBatch varCassetteBatch = new LdbCardBatch(
         this,
         argFrameworkBatch
      );

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardCassette.resurrectCassetteBatch()"
         );
      }
      return varCassetteBatch;
   }

   /********************************************************************
   * This method is called by the framework when a ComPoint object
   * receives a new message.  The cassette must use the connection
   * argument to receive and parse the message and create the
   * appropriate ProtocolRequest object for the message received.  The
   * framework will call the cassette's service() method passing the
   * returned ProtocolRequest object as an argument.
   * <p>
   * For this cassette, the cassette operations are always synchronous
   * and ComPoints are never used.  Therefore, if the framework calls
   * this method, a serious defect is indicated and the cassette
   * cannot continue.
   *
   * @param     argConnection - The connection for the ComPoint that
   * has received the new message.
   * @return    - The new ProtocolRequest that contains the relevant
   * data, parsed from the new message, to be sent to the cassette
   * for processing.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public ProtocolRequest createProtocolRequest(
      ETillConnection        argConnection
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardCassette.createProtocolRequest()"
         );
      }

      // Assertion.  This cassette does not use ComPoint objects.  If
      // the framework calls this method, there is a serious defect and
      // the cassette cannot continue.
      if (Trace.isAnyoneTracing()) {
         Trace.traceErrorOccurred(
            TRACE_ID,
            "LdbCardCassette The createProtocolRequest() method was called." +
            "\n\tPRC = " + PRC_CASSETTE_ERROR +
            "\n\tSRC = " + SRC_METHOD_NOT_IMPLEMENTED
         );
      }
      ErrorLog.logError(
         RB_ID,
         MSG_CASSETTE_METHOD_NOT_IMPLEMENTED,
         null,
         "LdbCardCassette.createProtocolRequest()"
      );
      Vector varVector = new Vector();
      varVector.addElement(
         "LdbCardCassette.createProtocolRequest()"
      );
      throw new ETillCassetteException(
         ErrorLog.lookupMessage(
            RB_ID,
            MSG_CASSETTE_METHOD_NOT_IMPLEMENTED,
            varVector
         )
      );
   }


   /********************************************************************
   * This method is used by the framework when the cassette is stopped.
   * It is the first call the framework makes to the cassette during
   * this sequence; it is called before the STOP_REQUEST is passed to
   * the service method.  A cassette uses this opportunity to destroy
   * all the ComPoint objects that were build during initialization.
   * <p>
   * For this cassette, ComPoint objects are not used so there
   * is nothing to do but return sucessfully.
   *
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public void destroyComPoints()
      throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardCassette.destroyComPoints()"
         );
      }


      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardCassette.destroyComPoints()"
         );
      }
   }

   /********************************************************************
   * This method is called by the framework when a the cassette is
   * being stopped.  This is the cassette's last opportunity to save any
   * intermediate results that will be recovered when the Payment
   * Manager is restarted.
   * <p>
   * For this cassette, the cassette operations are all synchronous and
   * nothing needs to be saved.  It merely logs a message to indicate
   * that the cassette has terminated.
   ********************************************************************/
   public void shutdown() {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardCassette.shutdown()"
         );
      }

      ErrorLog.logError(
         RB_ID,
         MSG_CASSETTE_HAS_STOPPED,
         null
      );

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardCassette.shutdown()"
         );
      }
   }

   /********************************************************************
   * This method is called by the framework to service administration
   * requests, order requests, batch requests, and protocol requests.
   * The cassette must determine the request type and route the
   * request to the correct object to be serviced.
   * <p>
   * For this cassette, the cassette examines the type of the request
   * using the instanceof operator.  Administration requests are sent to
   * the object affected by the request.  Order requests are sent to the
   * matching method in the cassette's order object.  Batch requests are
   * sent to the matching method in the cassette's batch object.  Any
   * unrecognized requests (like protocol requests, for example) will
   * result in an exception.
   *
   * @param     argRequest - the request object to be serviced
   * @param     argResponse - the response object to be updated with
   * the result of the service
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public void service(
      CassetteRequest        argRequest,
      CassetteResponse       argResponse
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardCassette.service()"
         );
         this.traceParameters(argRequest);
      }

      // Test if request is an order request.
      if (argRequest instanceof OrderRequest) {

         // Obtain the cassette's order object from the frame
         // order object which is referenced by the request.
         LdbCardOrder varCassetteOrder = (LdbCardOrder)
            (((OrderRequest)argRequest).getOrder().getCassetteOrder());

         // Test for each subclass of OrderRequest.  For each
         // subclass, call the matching method in the cassette
         // order object.  If the subclass is not recognized
         // throw an exception to abort the operation with the
         // appropriate return codes.
         if (argRequest instanceof ReceivePaymentRequest) {
            varCassetteOrder.receivePayment(
               (ReceivePaymentRequest)argRequest,
               (APIResponse)argResponse
            );
         } else if (argRequest instanceof AcceptPaymentRequest) {
            varCassetteOrder.acceptPayment(
               (AcceptPaymentRequest)argRequest,
               (APIResponse)argResponse
            );
         } else if (argRequest instanceof CancelOrderRequest) {
            varCassetteOrder.cancelOrder(
               (CancelOrderRequest)argRequest,
               (APIResponse)argResponse
            );
         } else if (argRequest instanceof CloseOrderRequest) {
            varCassetteOrder.closeOrder(
               (CloseOrderRequest)argRequest,
               (APIResponse)argResponse
            );
         } else if (argRequest instanceof ApproveRequest) {
            varCassetteOrder.approve(
               (ApproveRequest)argRequest,
               (APIResponse)argResponse
            );
         } else if (argRequest instanceof ApproveReversalRequest) {
            varCassetteOrder.approveReversal(
               (ApproveReversalRequest)argRequest,
               (APIResponse)argResponse
            );
         } else if (argRequest instanceof DepositRequest) {
            varCassetteOrder.deposit(
               (DepositRequest)argRequest,
               (APIResponse)argResponse
            );
         } else if (argRequest instanceof DepositReversalRequest) {
            varCassetteOrder.depositReversal(
               (DepositReversalRequest)argRequest,
               (APIResponse)argResponse
            );
         } else if (argRequest instanceof RefundRequest) {
            varCassetteOrder.refund(
               (RefundRequest)argRequest,
               (APIResponse)argResponse
            );
         } else if (argRequest instanceof RefundReversalRequest) {
            varCassetteOrder.refundReversal(
               (RefundReversalRequest)argRequest,
               (APIResponse)argResponse
            );
         } else {
            if (Trace.isAnyoneTracing()) {
               Trace.traceErrorOccurred(
                  TRACE_ID,
                  "LdbCardCassette Order request not recognized." +
                  "\n\tPRC = " + PRC_COMMAND_NOT_SUPPORTED +
                  "\n\tSRC = " + RC_NONE
               );
            }
            throw new ETillAbortOperation(
               PRC_COMMAND_NOT_SUPPORTED,
               RC_NONE
            );
         }

      // Test if request is a batch request.
      } else if (argRequest instanceof BatchRequest) {
         //Obtain the batch object from the batch request.
         LdbCardBatch varCassetteBatch = (LdbCardBatch)
            (((BatchRequest)argRequest).getBatch()).getCassetteBatch();
         //Test for each subclass of BatchRequest.  For each
         //subclass, call the matching method in the cassette batch
         // object. If the subclass is not recognized, throw an
         // exception.
         if (argRequest instanceof BatchOpenRequest) {
            varCassetteBatch.batchOpen(
               (BatchOpenRequest)argRequest,
               (APIResponse)argResponse
            );
         } else if (argRequest instanceof BatchCloseRequest) {
            varCassetteBatch.batchClose(
               (BatchCloseRequest)argRequest,
               (APIResponse)argResponse
            );
         } else if (argRequest instanceof BatchPurgeRequest) {
            varCassetteBatch.batchPurge(
               (BatchPurgeRequest)argRequest,
               (APIResponse)argResponse
            );
         } else if (argRequest instanceof DeleteBatchRequest) {
            varCassetteBatch.batchDelete(
               (DeleteBatchRequest)argRequest,
               (APIResponse)argResponse
            );
         } else {
            if (Trace.isAnyoneTracing()) {
               Trace.traceErrorOccurred(
                  TRACE_ID,
                  "LdbCardCassette Batch request not recognized." +
                  "\n\tPRC = " + PRC_COMMAND_NOT_SUPPORTED +
                  "\n\tSRC = " + RC_NONE
               );
            }
            throw new ETillAbortOperation(
               PRC_COMMAND_NOT_SUPPORTED,
               RC_NONE
            );
         }

      // Test if request is an administration request.
      } else if (argRequest instanceof AdminRequest) {

         this.serviceAdminRequest(
            (AdminRequest)argRequest,
            (APIResponse)argResponse
         );

      // The request type was not recognized, so throw an exception.
      } else {
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardCassette Request not recognized." +
               "\n\tPRC = " + PRC_COMMAND_NOT_SUPPORTED +
               "\n\tSRC = " + RC_NONE
            );
         }
         throw new ETillAbortOperation(
            PRC_COMMAND_NOT_SUPPORTED,
            RC_NONE
         );
      }

      // If there was any error detected, the cassette will throw
      // an exception and this logic will never be reached.  Since
      // all processing was successful, the service is committed.
      Supervisor.getThreadCommitPoint().commit();

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardCassette.service()"
         );
      }
   }

   /********************************************************************
   * This method is used by the cassette to trace the protocol data
   * keywords and values that were passed to the cassette.  Framework
   * keywords and values are expected to be traced by the framework.
   * <p>
   * Only keywords recognized by this cassette are included in the
   * trace.  Furthermore, password values are not displayed and
   * all but the first 4 and last 4 characters of the PAN are obscured
   * for security reasons.
   *
   * @param     argRequest - The request containing protocol data to
   * be traced.
   ********************************************************************/
   private void traceParameters(
      CassetteRequest argRequest
   ) {
      // Establish two buffers to hold trace data.
      StringBuffer varCommand = new StringBuffer(
         "\nLdbCard Service Request:  "
      );
      StringBuffer varParameters = new StringBuffer("\n");

      // If it is an APIRequest, find and trace the protocol data.
      if (argRequest instanceof APIRequest) {

         varCommand.append("APIRequest\n");
         ParameterTable varProtocolData =
            ((APIRequest)argRequest).getProtocolData();
         if (varProtocolData != null) {

            Enumeration varKeyList = varProtocolData.keys();
            while (varKeyList.hasMoreElements()) {
               String varKey = (String)varKeyList.nextElement();

               // Trace the protocol data only if it is recognized
               // by this cassette.
               if (clsValidationTable.containsKey(varKey)) {

                  // Obtain value as a string.
                  Object varValue = varProtocolData.get(varKey);

                  // Add keyword to trace buffer.
                  varParameters.append(varKey + "=");
                  if (varValue != null) {
                     if (varKey.equals(PD_ACC_PASSWORD)) {
                        //Password data is too sensitive to trace.
                        String varPassword = varValue.toString();
                        for (
                           int varIndex = 0;
                           varIndex < varPassword.length();
                           varIndex++
                        ) {
                           varParameters.append("*");
                        }
                     } else if (varKey.equals(PD_PAY_PAN)) {
                        // Except for the last four characters, the
                        // shopper's account number is obscured.
                        String varPan = varValue.toString();
                        int varBegIndex = varParameters.length();
                        varParameters.append(varPan);
                        int varEndIndex = varParameters.length() - 5;
                        for (
                           int varIndex = varBegIndex;
                           varIndex <= varEndIndex;
                           varIndex++
                        ) {
                           varParameters.setCharAt(varIndex,'*');
                        }
                     } else {
                        // In all other cases, the value is traced.
                        varParameters.append(varValue.toString());
                     }
                  }
                  varParameters.append("\n");
               }

            }

         }
      } else {
         varCommand.append("Unrecognized Request\n");
      }
      if (varParameters.length() < 2) {
         varParameters.append(
            "No cassette dependent protocol data was provided.\n"
         );
      }
      Trace.traceApiCommand(
         TRACE_ID,
         varCommand.toString(),
         varParameters.toString()
      );
   }

   /********************************************************************
   * This method is used by the cassette object to service
   * administration requests.  There are two cases:
   * <ol>
   * <li> The framework has asked the cassette to validate the
   * cassette dependent parameters before it starts its own
   * administration processing
   * <li> The framework has asked the cassette to process the
   * administration request and commit the resources affected.
   * </ol>
   * <p>
   * In other words, there are two cassette calls for every
   * administration request, one to validate the parameters and one to
   * perform the request.
   * <p>
   * For this cassette, this method checks to see if the request is
   * in validation mode and calls the correct method based on the
   * result.
   *
   * @param     argRequest - the administration request to be
   * serviced.
   * @param     argResponse - the response object to be updated with
   * the result of the service
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   private void serviceAdminRequest(
      AdminRequest           argRequest,
      APIResponse            argResponse
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardCassette.serviceAdminRequest()"
         );
      }

      // Check to see if the request is in validation mode.
      if (argRequest.validationMode()) {
         // Validate the cassette dependent parameters.
         this.validateAdminRequest(
            argRequest,
            argResponse
         );
      } else {
         // Process the request.
         this.processAdminRequest(
            argRequest,
            argResponse
         );
      }


      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardCassette.serviceAdminRequest()"
         );
      }
   }

   /********************************************************************
   * This method is used by the cassette object to validate the
   * cassette dependent data for any administration request.
   * <p>
   * For this cassette, there are only two administration requests that
   * could carry cassette dependent data:
   * <ol>
   * <li> Create Account
   * <li> Modify Account.
   * </ol>
   * @param     argRequest - the administration request to be validated.
   * @param     argResponse - the response object to be updated with
   * the result of the service
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   private void validateAdminRequest(
      AdminRequest           argRequest,
      APIResponse            argResponse
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardCassette.validateAdminRequest()"
         );
      }

      // Obtain token that identifies the administration request.
      switch (argRequest.getToken()) {

      // If the request creates a new account, the account class must
      // validate the protocol data.
      case CREATE_ACCOUNT_TOKEN:

         LdbCardAccount.validateCreateRequest(
            argRequest,
            argResponse
         );
         break;

      // If the request modifies an existing account, the account class
      // must validate the protocol data.
      case MODIFY_ACCOUNT_TOKEN:

         LdbCardAccount.validateModifyRequest(
            argRequest,
            argResponse
         );
         break;

      // In any remaining case, there is no protocol data to validate
      // so the validation is considered successful.
      default:
         argResponse.setReturnCodes(
            PRC_OPERATION_SUCCESS,
            RC_NONE
         );

      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardCassette.validateAdminRequest()"
         );
      }
   }

   /********************************************************************
   * This method is used by the cassette object to process any
   * administration requests.
   * <p>
   * For this cassette, there are only five administration requests that
   * require special processing:
   * <ol>
   * <li> Create Account
   * <li> Modify Account
   * <li> Delete Account
   * <li> Start Account
   * <li> Stop Account.
   * </ol>
   * @param     argRequest - the administration request to be processed.
   * @param     argResponse - the response object to be updated with
   * the result of the service.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   private void processAdminRequest(
      AdminRequest           argRequest,
      APIResponse            argResponse
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardCassette.processAdminRequest()"
         );
      }

      // Obtain token that identifies the administration request.
      switch (argRequest.getToken()) {

      // If the request creates a new account, construct a new
      // account object.
      case CREATE_ACCOUNT_TOKEN:
         {
            new LdbCardAccount(
               argRequest,
               argResponse
            );
         }
         break;

      // If the request modifies an existing account, find the
      // account object and ask it to modify itself.
      case MODIFY_ACCOUNT_TOKEN:

         {
            LdbCardAccount varAccount =
               LdbCardAccount.retrieveAccount(
                  (String)argRequest.getValue(MERCHANTNUMBER)
               );
            varAccount.modifyAccount(
               argRequest,
               argResponse
            );
         }
         break;

      // If the request deletes an existing account, find the
      // account object and ask it to delete itself.  Note: the
      // framework assures that there are no orders that reference
      // the account object before it allows the delete operation
      // to continue.
      case DELETE_ACCOUNT_TOKEN:

         {
            LdbCardAccount varAccount =
               LdbCardAccount.retrieveAccount(
                  (String)argRequest.getValue(MERCHANTNUMBER)
               );
            varAccount.deleteAccount(
               argRequest,
               argResponse
            );
         }
         break;

      // If the requests starts an existing account, find the
      // account object and ask it to start itself.
      case START_ACCOUNT_TOKEN:

         {
            LdbCardAccount varAccount =
               LdbCardAccount.retrieveAccount(
                  (String)argRequest.getValue(MERCHANTNUMBER)
               );
            varAccount.startAccount(
               argRequest,
               argResponse
            );
         }
         break;

      // If the request stops an existing account, find the
      // the account object and ask it to stop itself.
      case STOP_ACCOUNT_TOKEN:

         {
            LdbCardAccount varAccount =
               LdbCardAccount.retrieveAccount(
                  (String)argRequest.getValue(MERCHANTNUMBER)
               );
            varAccount.stopAccount(
               argRequest,
               argResponse
            );
         }
         break;


      // In any remaining case, there is nothing to do so the operation
      // is considered successful.
      default:
         argResponse.setReturnCodes(
            PRC_OPERATION_SUCCESS,
            RC_NONE
         );

      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardCassette.processAdminRequest()"
         );
      }
   }

}
