//**********************************************************************
// This source code is provided as an example.  No warranty is expressed
// or implied.  It is your responsibility to determine its suitability
// to your application, to modify it as necessary, and to validate
// its correctness.
//
// Licensed Materials - Property of IBM
// (c) Copyright IBM Corporation 2000. All Rights Reserved.
//
// US Government Users Restricted Rights - Use, duplication, or
// disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
//**********************************************************************
// Revision / History
// Defect     Date       Who      Description
// ---------- 01/03/2000 ldbader  Module created
//**********************************************************************

package com.ibm.etill.ldbcardcassette;


import  java.sql.Connection;
import  java.sql.PreparedStatement;
import  java.sql.ResultSet;
import  java.sql.SQLException;
import  java.util.Vector;

import  com.ibm.etill.framework.archive.ETillArchive;
import  com.ibm.etill.framework.cassette.APIResponse;
import  com.ibm.etill.framework.cassette.ApproveRequest;
import  com.ibm.etill.framework.cassette.ApproveReversalRequest;
import  com.ibm.etill.framework.cassette.Cassette;
import  com.ibm.etill.framework.cassette.CassetteTransaction;
import  com.ibm.etill.framework.cassette.DepositRequest;
import  com.ibm.etill.framework.cassette.DepositReversalRequest;
import  com.ibm.etill.framework.log.ErrorLog;
import  com.ibm.etill.framework.log.ETillCassetteException;
import  com.ibm.etill.framework.log.Trace;
import  com.ibm.etill.framework.payapi.Batch;
import  com.ibm.etill.framework.payapi.ETillAbortOperation;
import  com.ibm.etill.framework.payapi.Order;
import  com.ibm.etill.framework.payapi.ParameterTable;
import  com.ibm.etill.framework.payapi.Payment;
import  com.ibm.etill.framework.supervisor.Supervisor;

/***********************************************************************
* The cassette payment object is responsible for performing all the
* cassette's payment requests (Approve, ApproveReversal, Deposit,
* DepositReversal) and to manage the state of the framework
* payment object.
***********************************************************************/
public final class LdbCardPayment
   implements CassetteTransaction, LdbCardConstants {

   private static final String COPYRIGHT =
      "(c) Copyright IBM Corporation 2000 - All Rights Reserved.";

   //<gutter>
   //</gutter>

   //*******************************************************************
   //This is the cassette payment object's instance data.
   //******************************************************************/
   /********************************************************************
   * This cassette payment attribute contains the reference back to the
   * associated framework payment object.
   ********************************************************************/
   private Payment           objFrameworkPayment    = null;
   /********************************************************************
   * This cassette payment attribute contains the approval code.  The
   * approval code is supplied by the payment engine when a payment
   * transaction is successfully created and is used in subsequent
   * payment engine requests (like void or capture) to reference this
   * payment.
   ********************************************************************/
   private String            objApprovalCode        = null;
   /********************************************************************
   * This cassette payment attribute contains the decline reason. The
   * decline reason, originally small and therefore cryptic, was
   * designed to be displayed on a card swipe terminal at a point of
   * sale terminal.  In this cassette, it provides detailed information
   * about the cause of the failure.  (Since the text is provided by
   * the financial network, different for each network, and not
   * standardized in any way, it is impossible to translate these text
   * strings into return codes.)
   ********************************************************************/
   private String            objDeclineReason       = null;

   /********************************************************************
   * This constructor initializes a new cassette payment object.  A
   * reference to the associated framework's payment object is set.
   *
   * @param     argFrameworkPayment - The framework's payment object.
   * @param     argProtocolData - The protocol data parameter values
   *            provided by the merchant system.
   ********************************************************************/
   public LdbCardPayment(
      Payment         argFrameworkPayment,
      ParameterTable  argProtocolData
   ) {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardPayment.LdbCardPayment()"
         );
      }

      this.objFrameworkPayment  = argFrameworkPayment;

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardPayment.LdbCardPayment()"
         );
      }
   }

   /********************************************************************
   * This constructor initializes an existing cassette payment object.
   * A reference to the associated framework's payment object is set
   * and the remaining attributes are resurrected from the cassette's
   * payment database table.
   *
   * @param     argFrameworkPayment - The framework's payment object.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public LdbCardPayment(
      Payment         argFrameworkPayment
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardPayment.LdbCardPayment()"
         );
      }

      this.objFrameworkPayment  = argFrameworkPayment;
      this.resurrectPayment();

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardPayment.LdbCardPayment()"
         );
      }
   }

   /********************************************************************
   * This method is called by the framework when the merchant attempts
   * to Approve a payment that has previously been declined.  Usually
   * an Approve creates a new payment.  In this case, the payment exists
   * and must be reset to its initial state so that the Approval can
   * be attempted again.
   ********************************************************************/
   public void resetTransaction() {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardPayment.resetTransaction()"
         );
      }

      this.objApprovalCode  = null;
      this.objDeclineReason = null;

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardPayment.resetTransaction()"
         );
      }
   }

   /********************************************************************
   * This method is used by the cassette object to resurrect an
   * existing cassette payment from the cassette payment database table.
   * <p>
   * This method is <b>not</b> part of the Archivable interface.  The
   * methods for the Archivable interface are passed a JDBC database
   * connection for the operation.  In this case, an ETillArchive
   * object is constructed to provide a JDBC databaes connection.
   *
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   private void resurrectPayment() throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardPayment.resurrectPayment()"
         );
      }

      // The SQL select statement to obtain the cassette payment data.
      String varSelect =
         " SELECT * FROM " +
         ETillArchive.geteTillDatabaseOwner() + "." + TBL_PAYMENT +
         " WHERE " +
         PAY_MERCHANT + " = '" +
         this.objFrameworkPayment.getMerchantName() +
         "' AND " +
         PAY_ORDER    + " = '" +
         this.objFrameworkPayment.getOrderNumber() +
         "' AND " +
         PAY_PAYMENT  + " = '" +
         this.objFrameworkPayment.getTransactionNumber() +
         "'";

      // Construct a JDBC connection.
      ETillArchive varDatabase = new ETillArchive();

      // Lock the JDBC connection.
      synchronized (ETillArchive.getArchiveLock()) {
         try {
            // Trace database read.
            if (Trace.isAnyoneTracing()) {
               Trace.traceDatabaseRead(
                  TRACE_ID,
                  "\n" + varSelect
               );
            }
            // Execute the query.
            ResultSet varResult = varDatabase.beginQuery(varSelect);

            if (varResult.next()) {
               // The required data was found.  Initialize this
               // object's instance data.
               this.setApproval(
                  ETillArchive.readString(varResult, PAY_APPROVAL)
               );
               this.setReason(
                  ETillArchive.readString(varResult, PAY_REASON)
               );
               // Assertion test:  If there is more than one row in the
               // result set, there is a serious defect and the
               // cassette cannot continue.  There should never be more
               // than one row for a given cassette payment object.
               // The error is logged and a cassette exception is thrown.
               if (varResult.next()) {
                  if (Trace.isAnyoneTracing()) {
                     Trace.traceErrorOccurred(
                        TRACE_ID,
                        "LdbCardPayment Found too many rows." +
                        "\n\tPRC = " + PRC_DATABASE_ERROR +
                        "\n\tSRC = " + SRC_PAYMENT_SELECT_ROW_FAILURE
                     );
                  }
                  Vector varVector = new Vector();
                  varVector.addElement(
                     this.objFrameworkPayment.getMerchantName()
                  );
                  varVector.addElement(
                     this.objFrameworkPayment.getOrderNumber()
                  );
                  varVector.addElement(
                     this.objFrameworkPayment.getTransactionNumber()
                  );
                  ErrorLog.log(
                     RB_ID,
                     MSG_PAYMENT_SELECT_ROW_FAILURE,
                     varVector
                  );
                  throw new ETillCassetteException(
                     ErrorLog.lookupMessage(
                        RB_ID,
                        MSG_PAYMENT_SELECT_ROW_FAILURE,
                        varVector
                     )
                  );
               }
            } else {
               // The data was not found.  Log an error and abort.
               if (Trace.isAnyoneTracing()) {
                  Trace.traceErrorOccurred(
                     TRACE_ID,
                     "LdbCardPayment Resurrection failure." +
                     "\n\tPRC = " + PRC_DATABASE_ERROR +
                     "\n\tSRC = " + SRC_PAYMENT_SELECT_PAYMENT_MISSING
                  );
               }
               ErrorLog.logError(
                  RB_ID,
                  MSG_PAYMENT_SELECT_PAYMENT_MISSING,
                  null,
                  this.objFrameworkPayment.getMerchantName(),
                  this.objFrameworkPayment.getOrderNumber(),
                  this.objFrameworkPayment.getTransactionNumber()
               );
               throw new ETillAbortOperation(
                  PRC_DATABASE_ERROR,
                  SRC_PAYMENT_SELECT_PAYMENT_MISSING
               );
            }
         } catch (SQLException varException) {
            // The query failed.  Log the error and abort.
            if (Trace.isAnyoneTracing()) {
               Trace.traceErrorOccurred(
                  TRACE_ID,
                  "LdbCardPayment SQLException caught." +
                  "\n\tPRC = " + PRC_DATABASE_ERROR +
                  "\n\tSRC = " + SRC_PAYMENT_SELECT_SQL_FAILURE
               );
            }
            ErrorLog.logError(
               RB_ID,
               MSG_PAYMENT_SELECT_SQL_FAILURE,
               varException,
               this.objFrameworkPayment.getMerchantName(),
               this.objFrameworkPayment.getOrderNumber(),
               this.objFrameworkPayment.getTransactionNumber()
            );
            throw new ETillAbortOperation(
               PRC_DATABASE_ERROR,
               SRC_PAYMENT_SELECT_SQL_FAILURE
            );
         } finally {
            // Regardless of whether there was an exception or not,
            // clean up the ETillArchive object.
            try {
               varDatabase.endQuery();
            } catch (SQLException varException) {
               // If there was en error, it occured during cleanup so
               // there is no reason to abort.  Just log the error
               // because there is a defect in the ETillArchive object.
               if (Trace.isAnyoneTracing()) {
                  Trace.traceErrorOccurred(
                     TRACE_ID,
                     "LdbCardPayment SQL Exception caught." +
                     "\n\tPRC = " + PRC_DATABASE_ERROR +
                     "\n\tSRC = " + SRC_PAYMENT_SELECT_CLOSE_FAILURE
                  );
               }
               ErrorLog.logError(
                  RB_ID,
                  MSG_PAYMENT_SELECT_CLOSE_FAILURE,
                  varException,
                  this.objFrameworkPayment.getMerchantName(),
                  this.objFrameworkPayment.getOrderNumber(),
                  this.objFrameworkPayment.getTransactionNumber()
               );
            }
         }
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardPayment.resurrectPayment()"
         );
      }
   }

   /********************************************************************
   * This method is used by the framework order (a commitable object)
   * when its commit() method discovers the cassette payment object on
   * the framework order's create list.
   *
   * @param     argConnection - The JDBC database connection to be used
   * for this operation.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public void createRecord(
      Connection argConnection
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardPayment.createRecord()"
         );
      }
      try {
         // The SQL statement to insert a new cassette payment.
         String varInsert =
            " INSERT INTO " +
            ETillArchive.geteTillDatabaseOwner() + "." + TBL_PAYMENT + " (" +
            PAY_MERCHANT     + ", " +
            PAY_ORDER        + ", " +
            PAY_PAYMENT      + ", " +
            PAY_APPROVAL     + ", " +
            PAY_REASON       +
            ") VALUES (?, ?, ?, ?, ?)";
         if (Trace.isAnyoneTracing()) {
            Trace.traceDatabaseWrite(
               TRACE_ID,
               "\n" + varInsert + "\n" +
               PAY_MERCHANT  + " = " +
               this.objFrameworkPayment.getMerchantName() + "\n" +
               PAY_ORDER     + " = " +
               this.objFrameworkPayment.getOrderNumber() + "\n" +
               PAY_PAYMENT   + " = " +
               this.objFrameworkPayment.getTransactionNumber() + "\n" +
               PAY_APPROVAL  + " = " + this.getApproval() + "\n" +
               PAY_REASON    + " = " + this.getReason() + "\n"
            );
         }
         // Ask the connection to build a prepared statement using
         // the SQL statement defined above.
         PreparedStatement varStatement =
            argConnection.prepareStatement(varInsert);
         // Clear all the variable parameters from the statement.
         varStatement.clearParameters();
         // Tell the statement to set the variable values.
         ETillArchive.addString(
            this.objFrameworkPayment.getMerchantName(),  varStatement, 1
         );
         ETillArchive.addString(
            this.objFrameworkPayment.getOrderNumber(),   varStatement, 2
         );
         ETillArchive.addString(
            this.objFrameworkPayment.getTransactionNumber(),
                                                         varStatement, 3
         );
         ETillArchive.addString(
            this.getApproval(),                          varStatement, 4
         );
         ETillArchive.addString(
            this.getReason(),                            varStatement, 5
         );
         // Execute the insert.
         int varRowsAffected = varStatement.executeUpdate();
         // Close the statement.
         varStatement.close();
         // Assertion test:  If the number of affected rows is not 1
         // there is a serious defect and the cassette cannot continue.
         // The error is logged and a cassette exception is thrown.
         if (varRowsAffected != 1) {
            if (Trace.isAnyoneTracing()) {
               Trace.traceErrorOccurred(
                  TRACE_ID,
                  "LdbCardPayment Created too many rows." +
                  "\n\tPRC = " + PRC_DATABASE_ERROR +
                  "\n\tSRC = " + SRC_PAYMENT_CREATE_ROW_FAILURE
               );
            }
            Vector varVector = new Vector();
            varVector.addElement(
               this.objFrameworkPayment.getMerchantName()
            );
            varVector.addElement(
               this.objFrameworkPayment.getOrderNumber()
            );
            varVector.addElement(
               this.objFrameworkPayment.getTransactionNumber()
            );
            varVector.addElement(Integer.toString(varRowsAffected));
            ErrorLog.log(
               RB_ID,
               MSG_PAYMENT_CREATE_ROW_FAILURE,
               varVector
            );
            throw new ETillCassetteException(
               ErrorLog.lookupMessage(
                  RB_ID,
                  MSG_PAYMENT_CREATE_ROW_FAILURE,
                  varVector
               )
            );
         }
      }  catch (SQLException varException) {
         // The insert failed.  Log the error and abort.
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardPayment SQLException caught." +
               "\n\tPRC = " + PRC_DATABASE_ERROR +
               "\n\tSRC = " + SRC_PAYMENT_CREATE_SQL_FAILURE
            );
         }
         ErrorLog.logError(
            RB_ID,
            MSG_PAYMENT_CREATE_SQL_FAILURE,
            varException,
            this.objFrameworkPayment.getMerchantName(),
            this.objFrameworkPayment.getOrderNumber(),
            this.objFrameworkPayment.getTransactionNumber()
         );
         throw new ETillAbortOperation(
            PRC_DATABASE_ERROR,
            SRC_PAYMENT_CREATE_SQL_FAILURE
         );
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardPayment.createRecord()"
         );
      }
   }

   /********************************************************************
   * This method is used by the framework order (a commitable object)
   * when its commit() method discovers the cassette payment object on
   * the framework order's update list.
   *
   * @param     argConnection - The JDBC database connection to be used
   * for this operation.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public void updateRecord(
      Connection argConnection
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardPayment.updateRecord()"
         );
      }

      try {
         // The SQL statement to update an existing cassette payment.
         String varUpdate =
            " UPDATE " +
            ETillArchive.geteTillDatabaseOwner() + "." + TBL_PAYMENT + " SET " +
            PAY_APPROVAL     + "= ?, " +
            PAY_REASON       + "= ?  " +
            " WHERE " +
            PAY_MERCHANT + " = ? AND " +
            PAY_ORDER    + " = ? AND " +
            PAY_PAYMENT  + " = ?";
         if (Trace.isAnyoneTracing()) {
            Trace.traceDatabaseWrite(
               TRACE_ID,
               "\n" + varUpdate + "\n" +
               PAY_APPROVAL  + " = " + this.getApproval() + "\n" +
               PAY_REASON    + " = " + this.getReason() + "\n" +
               PAY_MERCHANT  + " = " +
               this.objFrameworkPayment.getMerchantName() + "\n" +
               PAY_ORDER     + " = " +
               this.objFrameworkPayment.getOrderNumber() + "\n" +
               PAY_PAYMENT   + " = " +
               this.objFrameworkPayment.getTransactionNumber() + "\n"
            );
         }
         // Ask the connection to build a prepared statement using
         // the SQL statement defined above.
         PreparedStatement varStatement =
            argConnection.prepareStatement(varUpdate);
         // Clear all the variable parameters from the statement.
         varStatement.clearParameters();
         // Tell the statement to set the variable values.
         ETillArchive.addString(
            this.getApproval(),                          varStatement, 1
         );
         ETillArchive.addString(
            this.getReason(),                            varStatement, 2
         );
         ETillArchive.addString(
            this.objFrameworkPayment.getMerchantName(),  varStatement, 3
         );
         ETillArchive.addString(
            this.objFrameworkPayment.getOrderNumber(),   varStatement, 4
         );
         ETillArchive.addString(
            this.objFrameworkPayment.getTransactionNumber(),
                                                         varStatement, 5
         );
         // Execute the update.
         int varRowsAffected = varStatement.executeUpdate();
         // Close the statement.
         varStatement.close();
         // Assertion test:  If the number of affected rows is not 1
         // there is a serious defect and the cassette cannot continue.
         // The error is logged and a cassette exception is thrown.
         if (varRowsAffected != 1) {
            if (Trace.isAnyoneTracing()) {
               Trace.traceErrorOccurred(
                  TRACE_ID,
                  "LdbCardPayment Modified too many rows." +
                  "\n\tPRC = " + PRC_DATABASE_ERROR +
                  "\n\tSRC = " + SRC_PAYMENT_UPDATE_ROW_FAILURE
               );
            }
            Vector varVector = new Vector();
            varVector.addElement(
               this.objFrameworkPayment.getMerchantName()
            );
            varVector.addElement(
               this.objFrameworkPayment.getOrderNumber()
            );
            varVector.addElement(
               this.objFrameworkPayment.getTransactionNumber()
            );
            varVector.addElement(Integer.toString(varRowsAffected));
            ErrorLog.log(
               RB_ID,
               MSG_PAYMENT_UPDATE_ROW_FAILURE,
               varVector
            );
            throw new ETillCassetteException(
               ErrorLog.lookupMessage(
                  RB_ID,
                  MSG_PAYMENT_UPDATE_ROW_FAILURE,
                  varVector
               )
            );
         }
      }  catch (SQLException varException) {
         // The update failed.  Log the error and abort.
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardPayment SQLException caught." +
               "\n\tPRC = " + PRC_DATABASE_ERROR +
               "\n\tSRC = " + SRC_PAYMENT_UPDATE_SQL_FAILURE
            );
         }
         ErrorLog.logError(
            RB_ID,
            MSG_PAYMENT_UPDATE_SQL_FAILURE,
            varException,
            this.objFrameworkPayment.getMerchantName(),
            this.objFrameworkPayment.getOrderNumber(),
            this.objFrameworkPayment.getTransactionNumber()
         );
         throw new ETillAbortOperation(
            PRC_DATABASE_ERROR,
            SRC_PAYMENT_UPDATE_SQL_FAILURE
         );
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardPayment.updateRecord()"
         );
      }
   }

   /********************************************************************
   * This method is used by the framework order (a commitable object)
   * when its commit() method discovers the cassette payment object on
   * the framework order's delete list.
   *
   * @param     argConnection - The JDBC database connection to be used
   * for this operation.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public void deleteRecord(
      Connection argConnection
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardPayment.deleteRecord()"
         );
      }

      try {
         // The SQL statement to delete an existing cassette payment.
         String varDelete =
            " DELETE FROM " +
            ETillArchive.geteTillDatabaseOwner() + "." + TBL_PAYMENT +
            " WHERE " +
            PAY_MERCHANT + " = ? AND " +
            PAY_ORDER    + " = ? AND " +
            PAY_PAYMENT  + " = ?";
         if (Trace.isAnyoneTracing()) {
            Trace.traceDatabaseWrite(
               TRACE_ID,
               "\n" + varDelete + "\n" +
               PAY_MERCHANT  + " = " +
               this.objFrameworkPayment.getMerchantName() + "\n" +
               PAY_ORDER     + " = " +
               this.objFrameworkPayment.getOrderNumber() + "\n" +
               PAY_PAYMENT   + " = " +
               this.objFrameworkPayment.getTransactionNumber() + "\n"
            );
         }
         // Ask the connection to build a prepared statement using
         // the SQL statement defined above.
         PreparedStatement varStatement =
            argConnection.prepareStatement(varDelete);
         // Clear all the variable parameters from the statement.
         varStatement.clearParameters();
         // Tell the statement to set the variable values.
         ETillArchive.addString(
            this.objFrameworkPayment.getMerchantName(),  varStatement, 1
         );
         ETillArchive.addString(
            this.objFrameworkPayment.getOrderNumber(),   varStatement, 2
         );
         ETillArchive.addString(
            this.objFrameworkPayment.getTransactionNumber(),
                                                         varStatement, 3
         );
         // Execute the delete.
         int varRowsAffected = varStatement.executeUpdate();
         // Close the statement.
         varStatement.close();
         // Assertion test:  If the number of affected rows is not 1
         // there is a serious defect and the cassette cannot continue.
         // The error is logged and a cassette exception is thrown.
         if (varRowsAffected != 1) {
            if (Trace.isAnyoneTracing()) {
               Trace.traceErrorOccurred(
                  TRACE_ID,
                  "LdbCardPayment Deleted too many rows." +
                  "\n\tPRC = " + PRC_DATABASE_ERROR +
                  "\n\tSRC = " + SRC_PAYMENT_DELETE_ROW_FAILURE
               );
            }
            Vector varVector = new Vector();
            varVector.addElement(
               this.objFrameworkPayment.getMerchantName()
            );
            varVector.addElement(
               this.objFrameworkPayment.getOrderNumber()
            );
            varVector.addElement(
               this.objFrameworkPayment.getTransactionNumber()
            );
            varVector.addElement(Integer.toString(varRowsAffected));
            ErrorLog.log(
               RB_ID,
               MSG_PAYMENT_DELETE_ROW_FAILURE,
               varVector
            );
            throw new ETillCassetteException(
               ErrorLog.lookupMessage(
                  RB_ID,
                  MSG_PAYMENT_DELETE_ROW_FAILURE,
                  varVector
               )
            );
         }
      }  catch (SQLException varException) {
         // The delete failed.  Log the error and abort.
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardPayment SQLException caught." +
               "\n\tPRC = " + PRC_DATABASE_ERROR +
               "\n\tSRC = " + SRC_PAYMENT_DELETE_SQL_FAILURE
            );
         }
         ErrorLog.logError(
            RB_ID,
            MSG_PAYMENT_DELETE_SQL_FAILURE,
            varException,
            this.objFrameworkPayment.getMerchantName(),
            this.objFrameworkPayment.getOrderNumber(),
            this.objFrameworkPayment.getTransactionNumber()
         );
         throw new ETillAbortOperation(
            PRC_DATABASE_ERROR,
            SRC_PAYMENT_DELETE_SQL_FAILURE
         );
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardPayment.deleteRecord()"
         );
      }
   }

   /********************************************************************
   * This method is used by the order object when it is asked to service
   * an Approve request.
   * <p>
   * For this cassette, the account object is used to authorize the
   * payment.  The account will be asked to authorize this payment and,
   * if successful and the merchant has requested an automatic deposit,
   * the account will be asked to capture this payment as well.  In
   * any case, the framework payment object's state is updated and
   * return codes are placed in the provided response object.
   *
   * @param     argRequest - An Approve order request.
   * @param     argResponse - The response object to hold the result.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public void approve(
      ApproveRequest         argRequest,
      APIResponse            argResponse
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardPayment.approve()"
         );
      }

      // Abort Approve if Payment is not reset, declined or void.
      if (
          this.objFrameworkPayment.getCurrentState() != PAYMENT_RESET     &&
          this.objFrameworkPayment.getCurrentState() != PAYMENT_DECLINED  &&
          this.objFrameworkPayment.getCurrentState() != PAYMENT_VOID
      ) {
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardPayment Approve was requested but the" +
               " payment is not in PAYMENT_RESET, PAYMENT_DECLINED, "+
               " or PAYMENT_VOID state." +
               "\n\tPRC    = " + PRC_VERB_NOT_VALID_IN_PRESENT_STATE +
               "\n\tSRC    = " + RC_PAYMENT +
               "\n\tState  = " + this.objFrameworkPayment.getCurrentState()
            );
         }
         throw new ETillAbortOperation (
            PRC_VERB_NOT_VALID_IN_PRESENT_STATE,
            RC_PAYMENT
         );
      }
      // Obtain the framework's order object from the request.
      Order varFrameworkOrder = argRequest.getOrder();
      // Obtain the cassette order object from the framework order
      // object.
      LdbCardOrder varCassetteOrder =
         (LdbCardOrder)varFrameworkOrder.getCassetteOrder();
      // Obtain the account object that is used to process
      // transactions for this order.
      LdbCardAccount varAccount = varCassetteOrder.getAccount();
      // Make sure that the approval amount will not exceed the total
      // order amount.
      this.checkApprovalAmount(varFrameworkOrder, varCassetteOrder);
      // Ask the account to authorize this payment.  If it returns
      // True, the authorization was approved.  Otherwise, the
      // authorization was declined.  The account will set this
      // object's approval code and decline reason before it returns.
      if (varAccount.authorizePayment(this)) {
         // Increment the total approved amount and add the
         // cassette order object to the update list.
         varCassetteOrder.setAmount(
            varCassetteOrder.getAmount() + this.objFrameworkPayment.getAmount()
         );
         Supervisor.getThreadCommitPoint().addToUpdateList(varFrameworkOrder);
         // It is the cassette's responsibility to set the payment
         // reference number to a value that the merchant can use to
         // identify the payment when reading a report sent by the
         // payment processor.
         this.objFrameworkPayment.setReferenceNumber(this.objApprovalCode);
         // The authorizaton was approved.  Update the state
         // information in the framework payment object and set
         // the correct return codes into the response.
         this.objFrameworkPayment.setCurrentState(PAYMENT_APPROVED);
         argResponse.setReturnCodes(
            PRC_OPERATION_SUCCESS,
            RC_NONE
         );
      } else {
         // The authorizaton was declined.  Update the state
         // information in the framework payment object and set
         // the correct return codes into the response.
         this.objFrameworkPayment.setCurrentState(PAYMENT_DECLINED);
         argResponse.setReturnCodes(
            PRC_FINANCIAL_FAILURE,
            RC_CASSETTE_DECLINED
         );
         // Trace the error.
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardPayment Payment authorization declined. " +
               "\n\tPRC = " + PRC_FINANCIAL_FAILURE +
               "\n\tSRC = " + RC_CASSETTE_DECLINED +
               "\n\tReason = " + this.getReason()
            );
         }
      }
      // There are two cases.  Either this is a new payment and it
      // must be added to the create list or it is an approval for
      // a payment that was previously declined and it must be added
      // to the update list.  The framework payment object contains
      // a boolean variable that can be tested to determine which
      // case this is.
      if (this.objFrameworkPayment.isNewTransaction())  {
         // Add the framework payment object to the framework
         // order's create list.  When the framework payment's
         // createRecord() method is called, it will call the
         // cassette payment's createRecord() method.
         Supervisor.getThreadCommitPoint().
            addToCreateList(this.objFrameworkPayment);
      } else {
         // Add the framework payment object to the framework
         // order's update list.  When the framework payment's
         // updateRecord() method is called, it will call the
         // cassette payment's updateRecord() method.
         Supervisor.getThreadCommitPoint().
            addToUpdateList(this.objFrameworkPayment);
      }
      this.objFrameworkPayment.setNewTransaction(false);
      // Check to see if the payment is approved.  If so, check to
      // see if the merchant requested an automatic deposit.
      if (
         (this.objFrameworkPayment.getCurrentState() == PAYMENT_APPROVED) &&
         argRequest.getDepositFlag()
      ) {
         // An automatic deposit is possible and has been requested.
         // Get the amount as an integer.
         int varAmount = argRequest.getAmount().getAmount();
         // Ask the account to capture this payment.  If it returns
         // True, the capture was approved.  Otherwise, the
         // capture was declined.  The account will set this
         // objects approval code and decline reason before it returns.
         if (varAccount.capturePayment(this, varAmount)) {
            // The capture was approved.  Update the framework object,
            // add the payment to the batch, and set return codes.
            // It is the cassette's responsibility to set the payment
            // reference number to a value that the merchant can use to
            // identify the payment when reading a report sent by the
            // payment processor.
            this.objFrameworkPayment.setReferenceNumber(this.objApprovalCode);
            // The framework payment has a special integer
            // that contains the amount that has been captured.  This
            // value is set here.
            this.objFrameworkPayment.setDepositAmount(varAmount);
            // The capture was approved.  Update the state
            // information in the framework payment object.
            this.objFrameworkPayment.setCurrentState(PAYMENT_DEPOSITED);
            // Since the capture was approved, add this payment to
            // the current batch and add the framework batch to the
            // commit point.
            varAccount.selectBatchAndAddPayment(this.objFrameworkPayment);
            // Set the return codes into the response.
            argResponse.setReturnCodes(
               PRC_OPERATION_SUCCESS,
               RC_NONE
            );
         } else {
            // The capture was declined.  Update the state
            // information in the framework payment object and set
            // the correct return codes into the response.
            this.objFrameworkPayment.setCurrentState(PAYMENT_APPROVED);
            argResponse.setReturnCodes(
               PRC_FINANCIAL_FAILURE,
               RC_CASSETTE_DECLINED
            );
            // Trace the error.
            if (Trace.isAnyoneTracing()) {
               Trace.traceErrorOccurred(
                  TRACE_ID,
                  "LdbCardPayment Payment capture declined. " +
                  "\n\tPRC = " + PRC_FINANCIAL_FAILURE +
                  "\n\tSRC = " + RC_CASSETTE_DECLINED +
                  "\n\tReason = " + this.getReason()
               );
            }
         }
         // Add the framework payment object to the framework
         // order's update list and commit.  When the framework
         // payment's updateRecord() method is called, it will
         // call the cassette's updateRecord() method.
         Supervisor.getThreadCommitPoint().
            addToUpdateList(this.objFrameworkPayment);
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardPayment.approve()"
         );
      }
   }

   /********************************************************************
   * This method is used when an Approve is requested.  It
   * checks to make sure that the new payment amount will not
   * increase the total approved amount to a value which exceeds the
   * amount of the order.
   *
   * @param     argFrameworkOrder - The Framework order object.
   * @param     argOrder - The cassette order object.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   private void checkApprovalAmount(
      Order       argFrameworkOrder,
      LdbCardOrder argOrder
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardPayment.checkApprovalAmount()"
         );
      }

      // Get the current amount from the framework payment.
      int varPaymentAmount = this.objFrameworkPayment.getAmount();
      // Get the total amount of all approved payments for this order;
      int varApprovedAmount = argOrder.getAmount();
      // Get the amount of the order.
      int varOrderAmount = argFrameworkOrder.getAmount().getAmount();
      // Abort the operation if the total amount of approved payments
      // plus the new amount proposed by the Approve request
      // exceeds the order's amount.
      if (varPaymentAmount + varApprovedAmount > varOrderAmount) {
         // This approval would exceed the total order amount.
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardPayment Approval amount too large." +
               "\n\tPRC = " + PRC_PARAMETER_VALUE_ERROR +
               "\n\tSRC = " + RC_AMOUNT
            );
         }
         throw new ETillAbortOperation(
            PRC_PARAMETER_VALUE_ERROR,
            RC_AMOUNT
         );
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardPayment.checkApprovalAmount()"
         );
      }
   }

   /********************************************************************
   * This method is used by the order object when it is asked to service
   * an ApproveReversal request.
   * <p>
   * For this cassette, the account object is used to void the previous
   * authorization and, if the new approval amount is not zero,
   * authorize again with the new amount.
   *
   * @param     argRequest - An ApproveReversal order request.
   * @param     argResponse - The response object to hold the result.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public void approveReversal(
      ApproveReversalRequest argRequest,
      APIResponse            argResponse
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardPayment.approveReversal()"
         );
      }

      // Abort ApproveReversal if payment is Payment is not approved
      if (this.objFrameworkPayment.getCurrentState() != PAYMENT_APPROVED) {
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardPayment ApproveReversal was requested but the" +
               " payment is not in PAYMENT_APPROVED state."+
               "\n\tPRC    = " + PRC_VERB_NOT_VALID_IN_PRESENT_STATE +
               "\n\tSRC    = " + RC_PAYMENT +
               "\n\tState  = " + this.objFrameworkPayment.getCurrentState()
            );
         }
         throw new ETillAbortOperation (
            PRC_VERB_NOT_VALID_IN_PRESENT_STATE,
            RC_PAYMENT
         );
      }
      // Obtain the framework's order object from the request.
      Order varFrameworkOrder = argRequest.getOrder();
      // Obtain the cassette order object from the framework order
      // object.
      LdbCardOrder varCassetteOrder =
         (LdbCardOrder)varFrameworkOrder.getCassetteOrder();
      // Obtain the account object that is used to process
      // transactions for this order.
      LdbCardAccount varAccount = varCassetteOrder.getAccount();
      // Check to make sure that the proposed amount is not too large.
      this.checkReversalAmount(
         varFrameworkOrder,
         varCassetteOrder,
         argRequest.getAmount().getAmount()
      );
      // Ask the account to void the last payment.  If it returns
      // True, the void was approved.  Otherwise, the void was
      // was declined.  The account will set this object's approval
      // code and decline reason before it returns.
      if (varAccount.voidPayment(this, this.objFrameworkPayment.getAmount())) {
         // Save the framework's current amount;
         int varOldAmount = this.objFrameworkPayment.getAmount();
         // Set the framework's amount to match the new approval
         // amount in the request.
         this.objFrameworkPayment.setAmount(
            argRequest.getAmount().getAmount()
         );
         // Check to see if the new amount is 0.
         if (argRequest.getAmount().getAmount() == 0) {
            // Update the total approved amount and add the
            // cassette order object to the update list.
            varCassetteOrder.setAmount(
               varCassetteOrder.getAmount() - varOldAmount
            );
            Supervisor.getThreadCommitPoint().addToUpdateList(varFrameworkOrder);
            // It is the cassette's responsibility to set the payment
            // reference number to a value that the merchant can use to
            // identify the payment when reading a report sent by the
            // payment processor.
            this.objFrameworkPayment.setReferenceNumber(this.objApprovalCode);
            // The new amount is 0 so set the framework payment state
            // to void and set the return codes in the response object.
            this.objFrameworkPayment.setCurrentState(PAYMENT_VOID);
            argResponse.setReturnCodes(
               PRC_OPERATION_SUCCESS,
               RC_NONE
            );
         } else {
            // The new amount is not zero.
            // Ask the account to authorize this payment.  If it returns
            // True, the authorization was approved.  Otherwise, the
            // authorization was declined.  The account will set this
            // object's approval code and decline reason before it returns.
            if (varAccount.authorizePayment(this)) {
               // Update the total approved amount and add the
               // cassette order object to the update list.
               varCassetteOrder.setAmount(
                  varCassetteOrder.getAmount() - varOldAmount +
                  argRequest.getAmount().getAmount()
               );
               Supervisor.getThreadCommitPoint().
                  addToUpdateList(varFrameworkOrder);
               // It is the cassette's responsibility to set the payment
               // reference number to a value that the merchant can use to
               // identify the payment when reading a report sent by the
               // payment processor.
               this.objFrameworkPayment.
                  setReferenceNumber(this.objApprovalCode);
               // The authorizaton was approved.  Update the state
               // information in the framework payment object and set
               // the correct return codes into the response.
               this.objFrameworkPayment.setCurrentState(PAYMENT_APPROVED);
               argResponse.setReturnCodes(
                  PRC_OPERATION_SUCCESS,
                  RC_NONE
               );
            } else {
               // The authorizaton was declined.  Update the state
               // information in the framework payment object and set
               // the correct return codes into the response.
               this.objFrameworkPayment.setCurrentState(PAYMENT_DECLINED);
               argResponse.setReturnCodes(
                  PRC_FINANCIAL_FAILURE,
                  RC_CASSETTE_DECLINED
               );
            }
         }
      } else {
         // The void was declined.  The framework payment object
         // should remain in Approved state, but the return codes
         // are set to indicate the failure.
         this.objFrameworkPayment.setCurrentState(PAYMENT_APPROVED);
         argResponse.setReturnCodes(
            PRC_FINANCIAL_FAILURE,
            RC_CASSETTE_DECLINED
         );
         // Trace the error.
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardPayment Payment void declined. " +
               "\n\tPRC = " + PRC_FINANCIAL_FAILURE +
               "\n\tSRC = " + RC_CASSETTE_DECLINED +
               "\n\tReason = " + this.getReason()
            );
         }
      }

      // Add the framework payment object to the framework
      // order's update list and commit.  When the framework
      // payment's updateRecord() method is called, it will
      // call the cassette's updateRecord() method.
      Supervisor.getThreadCommitPoint().
         addToUpdateList(this.objFrameworkPayment);

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardPayment.approveReversal()"
         );
      }
   }

   /********************************************************************
   * This method is used when an ApproveReversal is requested.  It
   * checks to make sure that the new payment amount will not
   * increase the total approved amount to a value which exceeds the
   * amount of the order.
   *
   * @param     argFrameworkOrder - The Framework order object.
   * @param     argOrder - The cassette order object.
   * @param     argNewAmount - The amount proposed by the request.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   private void checkReversalAmount(
      Order       argFrameworkOrder,
      LdbCardOrder argOrder,
      int         argNewAmount
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardPayment.checkReversalAmount()"
         );
      }

      // Get the current amount from the framework payment.
      int varPaymentAmount = this.objFrameworkPayment.getAmount();
      // Get the total amount of all approved payments for this order;
      int varApprovedAmount = argOrder.getAmount();
      // Get the amount of the order.
      int varOrderAmount = argFrameworkOrder.getAmount().getAmount();
      // Abort the operation if the total amount of approved payments
      // minus the original payment amount plus the new amount proposed
      // by the ApproveReversal request exceeds the order's amount.
      if (
         varApprovedAmount - varPaymentAmount + argNewAmount > varOrderAmount
      ) {
         // This approval would exceed the total order amount.
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardPayment Approval amount too large." +
               "\n\tPRC = " + PRC_PARAMETER_VALUE_ERROR +
               "\n\tSRC = " + RC_AMOUNT
            );
         }
         throw new ETillAbortOperation(
            PRC_PARAMETER_VALUE_ERROR,
            RC_AMOUNT
         );
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardPayment.checkReversalAmount()"
         );
      }
   }

   /********************************************************************
   * This method is used by the order object when it is asked to service
   * a Deposit request.
   * <p>
   * For this cassette, the account object is used to capture the
   * payment.
   *
   * @param     argRequest - A Deposit order request.
   * @param     argResponse - The response object to hold the result.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public void deposit(
      DepositRequest         argRequest,
      APIResponse            argResponse
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardPayment.deposit()"
         );
      }

      // Abort Deposit if payment is not approved
      if (this.objFrameworkPayment.getCurrentState() != PAYMENT_APPROVED) {
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardPayment Deposit was requested but the" +
               " payment is not in PAYMENT_APPROVED state."+
               "\n\tPRC    = " + PRC_VERB_NOT_VALID_IN_PRESENT_STATE +
               "\n\tSRC    = " + RC_PAYMENT +
               "\n\tState  = " + this.objFrameworkPayment.getCurrentState()
            );
         }
         throw new ETillAbortOperation (
            PRC_VERB_NOT_VALID_IN_PRESENT_STATE,
            RC_PAYMENT
         );
      }
      // Obtain the framework's order object from the request.
      Order varFrameworkOrder = argRequest.getOrder();
      // Obtain the cassette order object from the framework order
      // object.
      LdbCardOrder varCassetteOrder =
         (LdbCardOrder)varFrameworkOrder.getCassetteOrder();
      // Obtain the account object that is used to process
      // transactions for this order.
      LdbCardAccount varAccount = varCassetteOrder.getAccount();
      // Obtain the deposit amount from the request.
      int varAmount = argRequest.getAmount().getAmount();
      // Ask the account to capture this payment.  If it returns
      // True, the capture was approved.  Otherwise, the
      // capture was declined.  The account will set this
      // object's approval code and decline reason before it returns.
      if (varAccount.capturePayment(this, varAmount)) {
         // The capture was approved.  Update the framework object,
         // add the payment to the batch, and set return codes.
         // It is the cassette's responsibility to set the payment
         // reference number to a value that the merchant can use to
         // identify the payment when reading a report sent by the
         // payment processor.
         this.objFrameworkPayment.setReferenceNumber(this.objApprovalCode);
         // The framework payment has a special integer
         // that contains the amount that has been captured.  This
         // value is set here.
         this.objFrameworkPayment.setDepositAmount(varAmount);
         // The capture was approved.  Update the state
         // information in the framework payment object and set
         // the correct return codes into the response.
         this.objFrameworkPayment.setCurrentState(PAYMENT_DEPOSITED);
         // Since the capture was approved, add this payment to
         // the current batch and add the framework batch to the
         // commit point.
         varAccount.selectBatchAndAddPayment(this.objFrameworkPayment);
         argResponse.setReturnCodes(
            PRC_OPERATION_SUCCESS,
            RC_NONE
         );
      } else {
         // The capture was declined.  Update the state
         // information in the framework payment object and set
         // the correct return codes into the response.
         this.objFrameworkPayment.setCurrentState(PAYMENT_APPROVED);
         argResponse.setReturnCodes(
            PRC_FINANCIAL_FAILURE,
            RC_CASSETTE_DECLINED
         );
         // Trace the error.
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardPayment Payment capture declined. " +
               "\n\tPRC = " + PRC_FINANCIAL_FAILURE +
               "\n\tSRC = " + RC_CASSETTE_DECLINED +
               "\n\tReason = " + this.getReason()
            );
         }
      }
      // Add the framework payment object to the framework
      // order's update list and commit.  When the framework
      // payment's updateRecord() method is called, it will
      // call the cassette's updateRecord() method.
      Supervisor.getThreadCommitPoint().
         addToUpdateList(this.objFrameworkPayment);

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardPayment.deposit()"
         );
      }
   }

   /********************************************************************
   * This method is used by the order object when it is asked to service
   * a DepositReversal request.
   * <p>
   * For this cassette, the account is used to void
   * the previous payment and then authorize again with the previous
   * amount.  This is required because the framework expects the
   * payment to be in Approved state after processing a complete
   * deposit reversal but the payment engine voids the authorization
   * as well as the capture.
   *
   * @param     argRequest - A DepositReversal order request.
   * @param     argResponse - The response object to hold the result.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public void depositReversal(
      DepositReversalRequest argRequest,
      APIResponse            argResponse
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardPayment.depositReversal()"
         );
      }

      // Abort DepositReversal if payment is not deposited
      if (this.objFrameworkPayment.getCurrentState() != PAYMENT_DEPOSITED) {
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardPayment DepositReversal was requested but the" +
               " payment is not in PAYMENT_DEPOSITED state."+
               "\n\tPRC    = " + PRC_VERB_NOT_VALID_IN_PRESENT_STATE +
               "\n\tSRC    = " + RC_PAYMENT +
               "\n\tState  = " + this.objFrameworkPayment.getCurrentState()
            );
         }
         throw new ETillAbortOperation (
            PRC_VERB_NOT_VALID_IN_PRESENT_STATE,
            RC_PAYMENT
         );
      }
      // Obtain the framework's order object from the request.
      Order varFrameworkOrder = argRequest.getOrder();
      // Obtain the cassette order object from the framework order
      // object.
      LdbCardOrder varCassetteOrder =
         (LdbCardOrder)varFrameworkOrder.getCassetteOrder();
      // Obtain the account object that is used to process
      // transactions for this order.
      LdbCardAccount varAccount = varCassetteOrder.getAccount();
      // Ask the account to void the last payment.  If it returns
      // True, the void was approved.  Otherwise, the void was
      // was declined.  The account will set this object's approval
      // code and decline reason before it returns.
      if (
         varAccount.voidPayment(
            this,
            this.objFrameworkPayment.getDepositAmount()
         )
      ) {
         // The void was approved.  Remove the payment from the
         // batch, update the framework payment, and re-authorize.

         // Since the void was approved, remove this payment from
         // the current batch and add the framework batch to the
         // commit point.
         // Warning: the payment must be removed from the batch before
         // the framework payment object amount is set to zero.  In order
         // to update the batch payment totals correctly, the framework
         // batch subracts the amount in the framework payment object.
         varAccount.retrieveBatchAndRemovePayment(this.objFrameworkPayment);
         // The framework payment has a special integer that contains
         // the amount that has been captured.  This value is used in the
         // remove payment method, so it can't be set to zero until
         // after the payment has been removed from the batch.
         this.objFrameworkPayment.setDepositAmount(0);
         // Ask the account to authorize this payment again.  If it returns
         // True, the authorization was approved.  Otherwise, the
         // authorization was declined.  The account will set this
         // object's approval code and decline reason before it returns.
         if (varAccount.authorizePayment(this)) {
            // It is the cassette's responsibility to set the payment
            // reference number to a value that the merchant can use to
            // identify the payment when reading a report sent by the
            // payment processor.
            this.objFrameworkPayment.setReferenceNumber(this.objApprovalCode);
            // The authorizaton was approved.  Update the state
            // information in the framework payment object and set
            // the correct return codes into the response.
            this.objFrameworkPayment.setCurrentState(PAYMENT_APPROVED);
            argResponse.setReturnCodes(
               PRC_OPERATION_SUCCESS,
               RC_NONE
            );
         } else {
            // The authorizaton was declined.  Update the state
            // information in the framework payment object and set
            // the correct return codes into the response.
            this.objFrameworkPayment.setCurrentState(PAYMENT_DECLINED);
            argResponse.setReturnCodes(
               PRC_FINANCIAL_FAILURE,
               RC_CASSETTE_DECLINED
            );
            // Trace the error.
            if (Trace.isAnyoneTracing()) {
               Trace.traceErrorOccurred(
                  TRACE_ID,
                  "LdbCardPayment Payment authorization declined. " +
                  "\n\tPRC = " + PRC_FINANCIAL_FAILURE +
                  "\n\tSRC = " + RC_CASSETTE_DECLINED +
                  "\n\tReason = " + this.getReason()
               );
            }
         }
      } else {
         // The void was declined.  The framework payment object
         // should remain in Approved state, but the return codes
         // are set to indicate the failure.
         this.objFrameworkPayment.setCurrentState(PAYMENT_DEPOSITED);
         argResponse.setReturnCodes(
            PRC_FINANCIAL_FAILURE,
            RC_CASSETTE_DECLINED
         );
         // Trace the error.
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardPayment Payment void declined. " +
               "\n\tPRC = " + PRC_FINANCIAL_FAILURE +
               "\n\tSRC = " + RC_CASSETTE_DECLINED +
               "\n\tReason = " + this.getReason()
            );
         }
      }
      // Add the framework payment object to the framework
      // order's update list and commit.  When the framework
      // payment's updateRecord() method is called, it will
      // call the cassette credit's updateRecord() method.
      Supervisor.getThreadCommitPoint().
         addToUpdateList(this.objFrameworkPayment);

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardPayment.depositReversal()"
         );
      }
   }

   /********************************************************************
   * This method is used by the cassette order object when it is asked
   * to close a payment when a batch is being closed.
   ********************************************************************/
   public void closePayment() {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardPayment.closePayment()"
         );
      }

      // Assertion test.  If the payment is not in the correct state,
      // there is a serious defect and the cassette cannot continue.
      if (
         this.objFrameworkPayment.getCurrentState() != PAYMENT_DEPOSITED
      ) {
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardPayment ClosePayment was requested but the" +
               " payment is not in PAYMENT_DEPOSITED state."+
               "\n\tPRC    = " + PRC_VERB_NOT_VALID_IN_PRESENT_STATE +
               "\n\tSRC    = " + RC_PAYMENT +
               "\n\tState  = " + this.objFrameworkPayment.getCurrentState()
            );
         }
         Vector varVector = new Vector();
         varVector.addElement(
            this.objFrameworkPayment.getMerchantName()
         );
         varVector.addElement(
            this.objFrameworkPayment.getBatchNumber()
         );
         varVector.addElement(
            this.objFrameworkPayment.getOrderNumber()
         );
         varVector.addElement(
            this.objFrameworkPayment.getTransactionNumber()
         );
         varVector.addElement(
            Integer.toString(this.objFrameworkPayment.getCurrentState())
         );
         ErrorLog.log(
            RB_ID,
            MSG_PAYMENT_CLOSE_STATE_ERROR,
            varVector
         );
         throw new ETillCassetteException(
            ErrorLog.lookupMessage(
               RB_ID,
               MSG_PAYMENT_CLOSE_STATE_ERROR,
               varVector
            )
         );
      }

      // Set the payment state to closed and add the framework payment
      // to the update list.
      this.objFrameworkPayment.setCurrentState(PAYMENT_CLOSED);
      Supervisor.getThreadCommitPoint().
         addToUpdateList(this.objFrameworkPayment);

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardPayment.closePayment()"
         );
      }
   }

   /********************************************************************
   * This method is used to obtain the current approval code.
   *
   * @return    - The approval code returned by the payment engine.
   ********************************************************************/
   public String getApproval() {
      return this.objApprovalCode;
   }

   /********************************************************************
   * This method is used to obtain the framework's payment object.
   *
   * @return    - The framework's payment object for this payment.
   ********************************************************************/
   public Payment getFrameworkPayment() {
      return this.objFrameworkPayment;
   }

   /********************************************************************
   * This method is used to obtain the current reason code.
   *
   * @return    - The reason code (usually a text message) returned
   * by the payment engine when a transaction is declined.
   ********************************************************************/
   public String getReason() {
      return this.objDeclineReason;
   }

   /********************************************************************
   * This method is used to set the approval code for this payment.
   *
   * @param     argApprovalCode - The approval code returned by the
   * payment engine.
   ********************************************************************/
   public void setApproval(
      String argApprovalCode
   ) {
      this.objApprovalCode = argApprovalCode;
   }

   /********************************************************************
   * This method is used to set the decline reason for this payment.
   *
   * @param     argDeclineReason - The decline reason returned by the
   * payment engine.
   ********************************************************************/
   public void setReason(
      String argDeclineReason
   ) {
      this.objDeclineReason = argDeclineReason;
   }

}
