/*****************************************************************************
*
*  HTML Generator for Sample Store
*
*****************************************************************************/

import java.io.PrintWriter;

/****************************************************************************
 * NOTE: This is HTMLPage modified to accept a PrintWriter in the constructor
 * and use that PrintWriter to write to the browser.  The original HTMLPage
 * used System.out.println to write to the browser.
 *
 * HTMLPageSampleGUI generates HTML language constructs for delivery information to web
 * browsers.  The focus is on constructing dynamic web pages from database
 * information (hint) or other forms of personalization.  Minimal stylization
 * via abuse of presentation techniques is supported here, in favor of the new
 * "style sheet" technique.
 *
 * @author David M. Oliver (oliverdm@us.ibm.com)
 * @version 1.0 core functionality (only a few beautifying elements supported)
 * @version 1.1 style sheets, some beautification
 * @version 1.2 enhancements by Lance D Bader
 *****************************************************************************/

 public class HTMLPage
 {
    // **********************************************************************
    // **********************************************************************
    // instance variables:
    // I use these to keep (basic) track of stateful elements -- ones that
    // require start/end.  Similar elements that may be nested are not
    // included (caller is on his own)
    // **********************************************************************
    // **********************************************************************

    private int page_started, head_started, head_ended;
    private int body_started, form_started;
    private int list_indent;  // keep track of sub lists, so I can indent text
    private int table_indent; // same for tables

    public PrintWriter toClient;

    // **********************************************************************
    // **********************************************************************
    // class methods
    // **********************************************************************
    // **********************************************************************

    public HTMLPage(PrintWriter fromForm)
    {

        toClient = fromForm;

        this.page_started = 0;
        this.head_started = 0;
        this.body_started = 0;
        this.form_started = 0;
        this.head_ended   = 0;

        this.list_indent  = 0;
        this.table_indent = 0;
    }

    // **********************************************************************
    // **********************************************************************
    // instance methods
    // **********************************************************************
    // **********************************************************************

    /**
     * private methods
     *
     */

    private void PageSync()
    {
        if (page_started <= 0) StartPage();
        /***
        if ((head_started > 0) && (body_started <= 0)) EndHead();
        if ((head_ended > 0) && (body_started <= 0)) StartBody();
        if ((head_ended > 0) && (body_started > 0))
        {
            EndBody();
            EndPage();
        }
        ***/
    }

    private StringBuffer indent(int level)
    {
        int i;
        StringBuffer s = new StringBuffer();

        for (i = 0; i < level; i++) s.append("\t");

        return s;
    }

    // **********************************************************************
    // throw a random remark into the HTML page as a "comment"
    // **********************************************************************

    /**
      * Comment - place a sequence of text on the page as a (invisible) comment
      *
      * @param comment text of the comment
      */

    public void Comment(String comment)
    {
        toClient.println(indent(list_indent) + "\n<!-- " + comment + " -->\n");
    }

    /**
      * Comment - place a sequence of text on the page
      *
      * @param comment text of the comment
      */

    public void Text(String text)
    {
        toClient.print(text);
    }

    // **********************************************************************
    // basic start/end page, HEAD and BODY
    // **********************************************************************

    /**
      * StartPage - commence generation of a fresh HTML page
      *             (req'd at top of document)
      * @param none ()
      *
      */

    public void StartPage()
    {
        toClient.println("<HTML>");
        Comment("*** generated by Java class HTMLPage ***");
        toClient.println();

        page_started++;  // bump indicator
    }

    /**
     * EndPage - finish an HTML page
     *
     * @param none ()
     *
     */

    public void EndPage()
    {
        PageSync();

        toClient.println();
        Comment("*** thank you for using Java class HTMLPage! ***");
        toClient.println("</HTML>");

        page_started = 0;  // clear indicator
        head_ended = 0;
    }

    /**
      * StartHead - commence the HTML HEAD block
      *             (must appear before BODY block)
      * @param none ()
      *
      */

    public void StartHead()
    {
        PageSync();

        toClient.println("<HEAD>");
        toClient.println();

        head_started++;
    }

    /**
     * EndHead - end the HTML HEAD block
     *
     * @param none ()
     *
     */

    public void EndHead()
    {
        PageSync();

        toClient.println();
        toClient.println("</HEAD>");

        head_started = 0;
        head_ended++;
    }

    /**
      * StartBody - commence the HTML page body
      *             (appears after HTML HEAD)
      *
      * @param none ()
      *
      */

    public void StartBody()
    {
        PageSync();

        toClient.println();
        toClient.println("<BODY>");

        body_started++;
    }

    /**
     * StartBody - commence the HTML page body
     *             (appears after HTML HEAD)
     *
     * this version supports selection of the page background color
     *
     * @param bgcolor hex-format number specifying page background ("#fffcfc")
     */

    public void StartBody(String bgcolor)
    {
        PageSync();

        toClient.println();
        toClient.println("<BODY BGCOLOR=\"" + bgcolor + "\">");

        body_started++;
    }

    /**
     * StartBody - commence the HTML page body
     *             (appears after HTML HEAD)
     *
     * this version support selection of an image background, or alternative
     * solid color background
     *
     * @param bgcolor hex-format number specifying alt background color ("#fffcfc")
     * @param img_url URL pointing to image to use as background
     */

    public void StartBody(String bgcolor, String img_url)
    {
        PageSync();

        toClient.println();
        toClient.println("<BODY BGCOLOR=\"" + bgcolor +
                           "\" BACKGROUND=\"" + img_url + "\">");

        body_started++;
    }

    /**
     * EndBody - completes the HTML page body block
     *
     * @param none ()
     *
     */

    public void EndBody()
    {
        toClient.println();
        toClient.println("</BODY>");

        body_started = 0;
    }

    // **********************************************************************
    // The following elements are only allowed to appear in the HTML HEAD BLOCK.
    // **********************************************************************

    /**
     * Title - specify a title for the HTML page
     *
     * @param title title string
     */

    public void Title(String title)
    {
        toClient.println("<TITLE> " + title +  " </TITLE>");
    }

    /**
      * Base - specify the URL base to be used for URLs appearing on the page.
      *
      * @param baseURL the URL to be used as the base reference point
      *
      */

    public void Base(String baseURL)
    {
        toClient.println("<BASE HREF=\"" + baseURL + "\">");
    }

    /**
      * MetaName - supplies "meta" information about the HTML document
      *
      * @see HTMLPage#MetaEquiv
      * @param name meta-information name
      * @param content meta-information associated with this name
      *
      */

    public void MetaName(String name, String content)
    {
       toClient.println("<META " + "NAME=\"" + name + "\"" +
                          " CONTENT=\"" + content + "\">");
    }

    /**
     * MetaEquiv - supplies "meta" information, similar to HTTP MIME header
     *             information, about the document
     *
     * @see HTMLPage#MetaName
     * @param http_equiv meta-information name, equiv to HTTP header type
     * @param content the meta information for this HTTP header
     */

    public void MetaEquiv(String http_equiv, String content)
    {
        toClient.println("<META " + "HTTP-EQUIV=\"" + http_equiv + "\"" +
                           " CONTENT=\"" + content + "\">");
    }

    /**
     * StyleSheet - give reference to an external CSS1 style sheet
     *
     * @param sheet_url location of the style sheet
     */

    public void StyleSheet(String sheet_url)
    {
        toClient.println("<LINK REL=\"STYLESHEET\"" +
                           " TYPE=\"text/css\"" +
                           " HREF=\"" + sheet_url + "\">");
    }

    /**
     * Style - include a CSS1 style sheet
     *
     * @param style the style text
     */

    public void Style(String style)
    {
        toClient.println("<STYLE TYPE=\"text/css\">\n" +
                           "<!--\n" + style +
                           "\n  -->\n" +
                           "</STYLE>");
    }

    // **********************************************************************
    // The following elements are only allowed to appear in the HTML BODY BLOCK.
    // **********************************************************************

     /**
      * Script - element for including "scripting" code within the document
      *
      * @param language the scripting language of the contained script
      * @param the_script
      *
      */

     public void Script(String language, String the_script)
     {
        toClient.println("<SCRIPT LANGUAGE=\"" + language + "\">\n\n" +
                           the_script +  "\n\n</SCRIPT>");
     }

    // **********************************************************************
    // constructs that handle HTML/HTTP "forms"
    // **********************************************************************

     /**
      * StartForm - begins a "form"
      * This version allows only "POST"-type forms
      *
      * @param action URL of the CGI program that will handle this form
      *
      */

     public void StartForm(String action)
     {
        toClient.println();
        toClient.println("<FORM METHOD=\"POST\"" +
                           " ACTION=\"" + action + "\">"
                           );

        form_started++;
     }

     /**
      * EndForm - signify the end of the form
      *
      * @param none ()
      *
      */

     public void EndForm()
     {
        if (form_started > 0)
        {
            toClient.println();
            toClient.println("</FORM>");
        }

        form_started = 0;
     }

    // **********************************************************************
    // constructs that handle HTML tables
    // **********************************************************************

     /**
      * StartTable - commence a new HTML table, with default border
      *
      * @param none ()
      *
      */

     public void StartTable()
     {
        toClient.println("<TABLE BORDER=\"1\">");
        table_indent++;
     }

     /**
      * StartTable - commence a new HTML table, specifying border width
      *
      * this version allows selection of the border width
      *
      * @param border border present/not; if present, desired width
      *
      */

     public void StartTable(int border)
     {
        toClient.println("<TABLE BORDER=\"" + border + "\">");
        table_indent++;
     }

     /**
      * TableCaption - specify a Caption element for an open table
      *
      * @param cap the caption string
      *
      */

     public void TableCaption(String cap)
     {
        toClient.println(indent(table_indent) + "<CAPTION> " +
                           cap +  " </CAPTION>");
     }

     /**
      * TableCaption - specify a Caption element for an open table
      *
      * this version allows positional control of the element
      *
      * @param cap the caption string
      * @param where the location for the caption {top|bottom}
      *
      */

     public void TableCaption(String cap, String where)
     {
        toClient.println(indent(table_indent) +
                           "<CAPTION ALIGN=\"" + where + "\">" +
                           cap +  " </CAPTION>");
     }

     /**
      * StartTableRow - commence a new row within an open table
      *
      * @param none ()
      *
      */

     public void StartTableRow()
     {
        toClient.println(indent(table_indent) + "<TR>");
        table_indent++;
     }

     /**
      * StartTableRow - commence a new row within an open table, specifying
      *                 row alignment properties
      *
      * @param align the horizontal alignment of elements {left|center|right}
      * @param valign the vertical alignment of elements {top|middle|bottom|baseline}
      *
      */
     public void StartTableRow(String align, String valign)
     {
        toClient.println(indent(table_indent) + "<TR " +
                           "ALIGN=\"" + align + "\" VALIGN=\"" +
                           valign + "\">");
        table_indent++;
     }

     /**
      * TableHeader - specify a "header" element for this row (bold)
      *
      * @param header text string for header
      *
      */

     public void TableHeader(String header)
     {
        toClient.println(indent(table_indent) + "<TH> " +
                                  header + " </TH>");
     }

     /**
      * TableHeader - specify a "header" element for this row (bold), specifying
      *               alignment properties
      *
      * @param header text string for header
      * @param align the horizontal alignment of elements {left|center|right}
      * @param valign the vertical alignment of elements {top|middle|bottom|baseline}
      *
      */

     public void TableHeader(String header, String align, String valign)
     {
        toClient.println(indent(table_indent) + "<TH " +
                           "ALIGN=\"" + align + "\" VALIGN=\"" +
                           valign + "\"> " +
                           header + " </TH>");
     }

      /**
      * TableHeader - specify a "header" element for this row (bold), specifying
      *               alignment properties, and field width
      *
      * @param header text string for header
      * @param align the horizontal alignment of elements {left|center|right}
      * @param valign the vertical alignment of elements {top|middle|bottom|baseline}
      * @param width width of header within form row
      *
      */

     public void TableHeader(String header, String align, String valign, int width)
     {
        toClient.println(indent(table_indent) + "<TH " +
                           "WIDTH=\"" + width +
                           "\" ALIGN=\"" + align + "\" VALIGN=\"" +
                           valign + "\"> " +
                           header + " </TH>");
     }

     /**
      * TableData - indicate start of a data element within a row
      *
      * Use prior to inserting a FORM element into a row, or when data is
      * produced by another construct.  End with EndTableData().
      *
      * @param none ()
      *
      */

     public void TableData()
     {
        toClient.println(indent(table_indent) + "<TD> ");
        table_indent++;
     }

     /**
      * TableData - specify a data element within a row
      *
      * @param data data for this element
      *
      */

     public void TableData(String data)
     {
        toClient.println(indent(table_indent) + "<TD> " +
                                  data + " </TD>");
     }

     /**
      * TableData - specify a data element within a row, specifying
      *             alignment properties
      *
      * @param header text string for header
      * @param align the horizontal alignment of elements {left|center|right}
      * @param valign the vertical alignment of elements {top|middle|bottom|baseline}
      *
      */

     public void TableData(String data, String align, String valign)
     {
        toClient.println(indent(table_indent) + "<TD " +
                           "ALIGN=\"" + align + "\" VALIGN=\"" +
                           valign + "\"> " +
                           data + " </TD>");
     }

    /**
      * TableData - specify a data element within a row, specifying
      *               alignment properties, and field width
      *
      * @param header text string for header
      * @param align the horizontal alignment of elements {left|center|right}
      * @param valign the vertical alignment of elements {top|middle|bottom|baseline}
      * @param width the width of the data within the form row
      *
      */

     public void TableData(String data, String align, String valign, int width)
     {
        toClient.println(indent(table_indent) + "<TD " +
                           "WIDTH=\"" + width +
                           "\" ALIGN=\"" + align + "\" VALIGN=\"" +
                           valign + "\"> " +
                           data + " </TD>");
     }

     /**
      * TableParagraph - insert a paragraph break in table data
      *
      * @param none
      *
      */

     public void TableParagraph()
     {
        toClient.println(indent(table_indent) + "<P>");
     }

     /**
      * TableParagraph - insert a paragraph of text in table data
      *
      * @param ptext the text of the paragraph
      *
      */

     public void TableParagraph(String ptext)
     {
        toClient.println(indent(table_indent) + "<P> " + ptext +  "<P>");
     }

     /**
      * TableText - insert text in table data
      *
      * @param ptext the text
      *
      */

     public void TableText(String ptext)
     {
        toClient.println(indent(table_indent) + ptext);
     }

     /**
      * TableBreak - insert break in table data
      *
      */

     public void TableBreak()
     {
        toClient.println(indent(table_indent) + "<BR>");
     }

    /**
      * EndTableData - signal the end of a data element
      *
      * @param none ()
      *
      */

     public void EndTableData()
     {
        --table_indent;
        toClient.println(indent(table_indent) + "</TD>");
     }

     /**
      * EndTableRow - signal the end of header/data elements for this row
      *
      * @param none ()
      *
      */

     public void EndTableRow()
     {
        --table_indent;
        toClient.println(indent(table_indent) + "</TR>");
     }

      /**
      * EndTable - signify the end of an open table
      *
      * @param none ()
      *
      */

     public void EndTable()
     {
        --table_indent;
        toClient.println("</TABLE>");
        if (table_indent > 0)
        {
            table_indent = 0;
            Comment("I just had to force table indent to zero");
        }
     }

    // **********************************************************************
    // constructs that handle HTML selectors of many types
    // **********************************************************************

     /**
      * StartSelector - start a multi-item value selector("menu" type)
      *
      * @param name name, within the form, to use for this selector
      *
      */

     public void StartSelector(String name)
     {
        toClient.println(indent(table_indent) + "<SELECT NAME=\"" + name + "\">");
        list_indent++;
     }

     /**
      * StartSelector - start a multi-item value selector("menu" type), with size
      *
      * @param name name, within the form, to use for this selector
      * @param size the number of items that should show when displayed
      *
      */

     public void StartSelector(String name, int size)
     {
        toClient.println(indent(table_indent) +
                           "<SELECT NAME=\"" + name + "\" SIZE=\"" + size + "\">");
        list_indent++;
     }

     /**
      * StartSelector - start a multi-item value selector("menu" type), with size
      * and option of multi-selection
      *
      * @param name name, within the form, to use for this selector
      * @param size the number of items that should show when displayed
      * @param multiple allowmultiple selections
      *
      */

     public void StartSelector(String name, int size, int multiple)
     {
        toClient.println(indent(table_indent) +
                           "<SELECT NAME=\"" + name + "\" SIZE=\"" + size +
                           "\" MULTIPLE>");
        list_indent++;
     }

     /**
      * SelectorOption - specifies one of the choices displayed in a Selector
      *                  This form allows pre-selection of an entry, and the use
      *                  of an option "value" different from the option "name"
      *
      * @param option text string to show as the option
      * @param selected 0/1 to pre-select an entry
      * @param value the value of the option
      */

     public void SelectorOption(String option, int selected, String value)
     {
        StringBuffer sb = new StringBuffer();

        sb.append(indent(table_indent));
        sb.append(indent(list_indent));
        sb.append("<OPTION VALUE=\"" + value + "\"");
        if (selected > 0) sb.append(" SELECTED");
        sb.append("> " + option);

        toClient.println(sb);
     }

    /**
      * SelectorOption - specifies one of the choices displayed in a Selector
      *
      * @param option text string to show as the option
      *
      */

     public void SelectorOption(String option)
     {
        SelectorOption(option, 0, option);
     }

     /**
      * SelectorOption - specifies one of the choices displayed in a Selector
      *                  This form allows pre-selection of an entry
      *
      * @param option text string to show as the option
      * @param selected 0/1 to pre-select an entry
      *
      */

     public void SelectorOption(String option, int selected)
     {
        SelectorOption(option, selected, option);
     }

     /**
      * EndSelector - end a multi-item value selector("menu" type)
      *
      * @param none ()
      *
      */

     public void EndSelector()
     {
        list_indent--;
        toClient.println(indent(table_indent) + "</SELECT>");
     }

     /**
      * CheckBox - an element within a group of buttons -- ALL with same name --
      *            that allows multiple selections from a list of options)
      *
      * @param name name to be used, within the form, for this option
      * @param value value for the option
      *
      */

     public void CheckBox(String name, String value)
     {
        toClient.println(indent(table_indent) +
                           "<INPUT TYPE=\"CHECKBOX\" NAME=\"" + name +
                           "\" VALUE=\"" + value + "\">" + value);
     }

    /**
      * CheckBox - an element within a group of buttons -- ALL with same name --
      *            that allows multiple selections from a list of options). This
      *            form allows options to be pre-selected
      *
      * @param name name to be used, within the form, for this option
      * @param value value for the option
      * @param selected 0/1 if pre-selected
      *
      */

     public void CheckBox(String name, String value, int selected)
     {
        StringBuffer sb = new StringBuffer();

        sb.append(indent(table_indent));
        sb.append("<INPUT TYPE=\"CHECKBOX\" NAME=\"" + name + "\" VALUE=\"" + value + "\"");
        if (selected > 0) sb.append(" CHECKED");
        sb.append("> " + value);

        toClient.println(sb);
     }

     /**
      * CheckBox - an element within a group of buttons -- ALL with same name --
      *            that allows multiple selections from a list of options). This
      *            form allows options to be pre-selected and appear with/out label
      *
      * @param name name to be used, within the form, for this option
      * @param value value for the option
      * @param selected 0/1 if pre-selected
      * @param label 0/1 if label is to be shown aside Box
      */

     public void CheckBox(String name, String value, int selected, int label)
     {
        StringBuffer sb = new StringBuffer();

        sb.append(indent(table_indent));
        sb.append("<INPUT TYPE=\"CHECKBOX\" NAME=\"" + name + "\" VALUE=\"" + value + "\"");
        if (selected > 0) sb.append(" CHECKED");
        sb.append(">");
        if (label > 0) sb.append(" " + value);

        toClient.println(sb);
     }

     /**
      * ImageButton - a graphical toggle button
      *
      * @param name name to be used, within the form, for this option
      * @param value value for the option
      * @param source the URL location of the source imagery
      *
      */

     public void ImageButton(String name, String value, String source)
     {
        toClient.println(indent(table_indent) +
                           "<INPUT TYPE=\"IMAGE\" BORDER=\"0\" " +
                           "SRC=\"" + source + "\" NAME=\"" + name +
                           "\" VALUE=\"" + value + "\">" + value);
     }

     /**
      * RadioButton - an element with a group of buttons -- ALL with same name --
      *               that allows a single selection from a list of options)
      *
      * @param name name to be used, within the form, for this option
      * @param value value for the option
      *
      */

     public void RadioButton(String name, String value)
     {
        toClient.println(indent(table_indent) +
                           "<INPUT TYPE=\"RADIO\" NAME=\"" + name +
                           "\" VALUE=\"" + value + "\">" + value);
     }

     /**
      * RadioButton - an element with a group of buttons -- ALL with same name --
      *               that allows a single selection from a list of options)
      *
      * @param name name to be used, within the form, for this option
      * @param value value for the option
      * @param label lable for the option
      *
      */

     public void RadioButton(String name, String value, String label)
     {
        toClient.println(indent(table_indent) +
                           "<INPUT TYPE=\"RADIO\" NAME=\"" + name +
                           "\" VALUE=\"" + value + "\">" + label);
     }

    /**
      * RadioButton - an element with a group of buttons -- ALL with same name --
      *               that allows a single selection from a list of options). This
      *               form allows an item to be pre-selected
      *
      * @param name name to be used, within the form, for this option
      * @param value value for the option
      * @param selected 0/1 pre-select an item
      *
      */

     public void RadioButton(String name, String value, int selected)
     {
        StringBuffer sb = new StringBuffer();

        sb.append(indent(table_indent));
        sb.append("<INPUT TYPE=\"RADIO\" NAME=\"" + name + "\" VALUE=\"" + value + "\"");
        if (selected > 0) sb.append(" CHECKED");
        sb.append("> " + value);

        toClient.println(sb);
     }

    /**
      * RadioButton - an element with a group of buttons -- ALL with same name --
      *               that allows a single selection from a list of options). This
      *               form allows a label and an item to be pre-selected
      *
      * @param name name to be used, within the form, for this option
      * @param value value for the option
      * @param label value for the option
      * @param selected 0/1 pre-select an item
      *
      */

     public void RadioButton(String name, String value, String label, int selected)
     {
        StringBuffer sb = new StringBuffer();

        sb.append(indent(table_indent));
        sb.append("<INPUT TYPE=\"RADIO\" NAME=\"" + name + "\" VALUE=\"" + value + "\"");
        if (selected > 0) sb.append(" CHECKED");
        sb.append("> " + label);

        toClient.println(sb);
     }

    /**
      * RadioButton - an element with a group of buttons -- ALL with same name --
      *               that allows a single selection from a list of options). This
      *               form allows an item to be pre-selected
      *
      * @param name name to be used, within the form, for this option
      * @param value value for the option
      * @param selected 0/1 pre-select an item
      * @param label 0/1 if label is to be shown aside Button
      */

     public void RadioButton(String name, String value, int selected, int label)
     {
        StringBuffer sb = new StringBuffer();

        sb.append(indent(table_indent));
        sb.append("<INPUT TYPE=\"RADIO\" NAME=\"" + name + "\" VALUE=\"" + value + "\"");
        if (selected > 0) sb.append(" CHECKED");
        sb.append(">");
        if (label > 0) sb.append(" " + value);

        toClient.println(sb);
     }

    // **********************************************************************
    // constructs that handle miscellaneous FORM elements
    // **********************************************************************

     /**
      * HiddenField - place a hidden name/value pair in the form
      *
      * @param name name to be used, within the form, for this field
      * @param value value for the field
      *
      */

     public void HiddenField(String name, String value)
     {
        toClient.println(indent(table_indent) +
                           "<INPUT TYPE=\"HIDDEN\" NAME=\"" + name +
                           "\" VALUE=\"" + value + "\">");
     }

    /**
      * PasswordField - place a Password-input entry on the page
      *
      * @param name name to be used, within the form, for this field
      * @param size width of the input, in chars
      * @param max the maximum number of characters that can be entered
      *
      */

     public void PasswordField(String name, int size, int max)
     {
        toClient.println(indent(table_indent) +
                           "<INPUT TYPE=\"PASSWORD\" NAME=\"" + name +
                           "\" SIZE=\"" + size +
                           "\" MAXLENGTH=\"" + max + "\">");
     }

    /**
      * PasswordField - place a Password-input entry on the page
      *
      * @param name name to be used, within the form, for this field
      * @param size width of the input, in chars
      * @param max the maximum number of characters that can be entered
      * @param predef a predefined value for the field
      *
      */

     public void PasswordField(String name, int size, int max, String predef)
     {
        toClient.println(indent(table_indent) +
                           "<INPUT TYPE=\"PASSWORD\" NAME=\"" + name +
                           "\" SIZE=\"" + size +
                           "\" MAXLENGTH=\"" + max +
                           "\" VALUE=\"" + predef + "\">");
     }

     /**
      * TextField - place a text-input entry on the page
      *
      * @param name name to be used, within the form, for this field
      * @param size width of the input, in chars
      * @param max the maximum number of characters that can be entered
      *
      */

     public void TextField(String name, int size, int max)
     {
        toClient.println(indent(table_indent) +
                           "<INPUT TYPE=\"TEXT\" NAME=\"" + name +
                           "\" SIZE=\"" + size +
                           "\" MAXLENGTH=\"" + max + "\">");
     }

    /**
      * TextField - place a text-input entry on the page
      *
      * @param name name to be used, within the form, for this field
      * @param size width of the input, in chars
      * @param max the maximum number of characters that can be entered
      * @param predef a predefined value for the field
      *
      */

     public void TextField(String name, int size, int max, String predef)
     {
        toClient.print(indent(table_indent) +
                           "<INPUT TYPE=\"TEXT\" NAME=\"" + name +
                           "\" SIZE=\"" + size +
                           "\" MAXLENGTH=\"" + max);
        if (predef != null) {
           toClient.print("\" VALUE=\"" + predef );
        }
        toClient.println("\">");
     }

     /**
      * NumberField - place a numeric-input entry on the page
      *
      * @param name name to be used, within the form, for this field
      * @param size width of the input, in chars
      * @param max the maximum number of characters that can be entered
      *
      */

     public void NumberField(String name, int size, int max)
     {
        toClient.println(indent(table_indent) +
                           "<INPUT TYPE=\"NUMBER\" NAME=\"" + name +
                           "\" SIZE=\"" + size +
                           "\" MAXLENGTH=\"" + max + "\">");
     }

     /**
      * NumberField - place a numeric-input entry on the page
      *
      * @param name name to be used, within the form, for this field
      * @param size width of the input, in chars
      * @param max the maximum number of characters that can be entered
      * @param predef a predefined value for the field
      *
      */

     public void NumberField(String name, int size, int max, String predef)
     {
        toClient.println(indent(table_indent) +
                           "<INPUT TYPE=\"NUMBER\" NAME=\"" + name +
                           "\" SIZE=\"" + size +
                           "\" MAXLENGTH=\"" + max +
                           "\" VALUE=\"" + predef + "\">");
     }

     /**
      * SubmitButton - place a SUBMIT-type action button in the form
      *
      * @param name name to be used, within the form, for this field
      * @param value value for the field
      *
      */

     public void SubmitButton(String name, String value)
     {
        toClient.println(indent(table_indent) +
                           "<INPUT TYPE=\"SUBMIT\" NAME=\"" + name +
                           "\" VALUE=\"" + value + "\">");
     }

     /**
      * ResetButton - place a RESET-type action button on the page
      *
      * @param name name to be used, within the form, for this field
      * @param value value for the field
      *
      */

     public void ResetButton(String name, String value)
     {
        toClient.println(indent(table_indent) +
                           "<INPUT TYPE=\"RESET\" NAME=\"" + name +
                           "\" VALUE=\"" + value + "\">");
     }


    // **********************************************************************
    // constructs that handle HTML lists of many types
    // **********************************************************************

     /**
      * StartGlossary - commence a Glossary-type list
      *
      * @param none ()
      *
      */

     public void StartGlossary()
     {
        toClient.println(indent(list_indent) + "<DL>");

        list_indent++;
     }

     /**
      * StartGlossary - commence a Glossary-type list
      *
      * @param compact indicate if this list should be display in compact form
      *
      */

     public void StartGlossary(int compact)
     {
        toClient.println(indent(list_indent) + "<DL COMPACT>");

        list_indent++;
     }

     /**
      * GlossaryItem - the "name" of an element within a GlossaryList
      *
      * @param item item name
      *
      */

     public void GlossaryItem(String item)
     {
        toClient.println(indent(list_indent) + "<DT> " + item +  "</DT>");
     }

     /**
      * GlossaryDescription - the "value" of an element within a GlossaryList
      *
      * @param text the description of a GlossaryItem
      *
      */

     public void GlossaryDescription(String text)
     {
        toClient.println(indent(list_indent) + "<DD> " + text +  "</DD>");
     }

     /**
      * EndGlossary - close up a GlossaryList
      *
      * @param none
      *
      */

     public void EndGlossary()
     {
        list_indent--;
        toClient.println(indent(list_indent) + "</DL>");
     }

     /**
      * StartList - commence a generic list
      *
      * @param none ()
      *
      */

     public void StartList()
     {
        toClient.println(indent(list_indent) + "<UL>");

        list_indent++;
     }

     /**
      * StartList - commence a generic list, opt displayed in compact form
      *
      * @param compact 1|0 display compact
      *
      */

     public void StartList(int compact)
     {
        if (compact > 0) toClient.println(indent(list_indent) + "<UL COMPACT>");
        else toClient.println(indent(list_indent) + "<UL>");

        list_indent++;
     }

     /**
      * StartList - commence a list, opt displayed in compact form
      *
      * @param type type of list to display {Ordered|Unordered}
      * @param compact 1|0 display compact
      *
      */

     public void StartList(String type, int compact)
     {
        String st;

        if (type.equals("Ordered")) st = new String("OL");
        else st = new String("UL");

        if (compact > 0) toClient.println(indent(list_indent) + "<" + st + " COMPACT>");
        else toClient.println(indent(list_indent) + "<" + st + ">");

        list_indent++;
     }

     /**
      * ListItem - an entry within an open list
      *
      * @param item text for the list item
      *
      */

     public void ListItem(String item)
     {
        toClient.println(indent(list_indent) + "<LI> " + item);
     }

     /**
      * EndList - close an open (generic) list
      *
      * @param none ()
      *
      */

     public void EndList()
     {
        --list_indent;
        toClient.println(indent(list_indent) + "</UL>");
     }

     /**
      * EndList - close an open list of a specified type
      *
      * @param type type of list to close {Ordered|Unordered}
      *
      */

     public void EndList(String type)
     {
        String st;

        if (type.equals("Ordered")) st = new String("OL");
        else st = new String("UL");

        --list_indent;
        toClient.println(indent(list_indent) + "</" + st + ">");
     }

     // **********************************************************************
     // constructs that handle HTML Hypertext references
     // **********************************************************************

     /**
      * TextLink - create a hypertext link
      *
      * @param text the visible string shown
      * @param link the link URL
      */

     public void TextLink(String text, String URL) {
        toClient.println("<A HREF=\"" + URL + "\">" + text + "</A>");
     }

     /**
      * ImageLink - create a hyper image link
      *
      * @param image the URL of the image source
      * @param link the link URL
      */

     public void ImageLink(String source, String URL) {
        toClient.println("<A HREF=\"" + URL + "\">");
        toClient.println("<IMG BORDER=0 ALIGN=\"CENTER\" ALT=\"image\"" +
                           " SRC=\"" + source + "\"></A>");
     }

     /**
      * ImageLink - create a hyper image link
      *
      * @param image the URL of the image source
      * @param link the link URL
      * @param align image alignment (RIGHT, LEFT, CENTER)
      */

     public void ImageLink(String source, String URL, String align) {
        toClient.println("<A HREF=\"" + URL + "\">");
        toClient.println("<IMG BORDER=0 ALIGN=\"" + align +
                           "\" ALT=\"image\"" +
                           " SRC=\"" + source + "\"></A>");
     }

     /**
      * ImageLink - create a hyper image link
      *
      * @param image the URL of the image source
      * @param link the link URL
      * @param align image alignment (RIGHT, LEFT, CENTER)
      * @param alt alternative text to display
      */

     public void ImageLink(String source, String URL, String align, String alt) {
        toClient.println("<A HREF=\"" + URL + "\">");
        toClient.println("<IMG BORDER=0 ALIGN=\"" + align +
                           "\" ALT=\"" + alt + "\"" +
                           " SRC=\"" + source + "\"></A>");
     }

     // **********************************************************************
     // constructs that handle HTML Imagery and client-side image maps
     // **********************************************************************

     /**
      * Image - display an image
      *
      * @param source the URL location of the image
      * @param alt the alternative text to display
      */

     public void Image(String source, String alt) {
        toClient.println("IMG SRC=\"" + source + "\" ALT=\"" + alt + "\">");
     }

     /**
      * Image - display an image
      *
      * @param source the URL location of the image
      * @param alt the alternative text to display
      * @param width size
      * @param height size
      */

     public void Image(String source, String alt, int width, int height) {
        toClient.println("IMG SRC=\"" + source + "\" ALT=\"" + alt +
                           "\" WIDTH=\"" + width + "\" HEIGHT=\"" +
                           height + "\">");
     }

     /**
      * ImageMap - start display of an image map
      *
      * @param source the URL location of the image
      * @param alt the alternative text to display
      * @param width size
      * @param height size
      * @param mapname name of the image map to use
      */

     public void ImageMap(String source, String alt, int width, int height, String mapname) {
        toClient.println("IMG SRC=\"" + source + "\" ALT=\"" + alt +
                           "\" WIDTH=\"" + width + "\" HEIGHT=\"" +
                           height + "\" USEMAP=\"#" + mapname + "\">");
     }

     /**
      * MapSpec - start an image map specification
      *
      * @param mapname name of this map
      */

     public void MapSpec(String mapname) {
        toClient.println("<MAP NAME=\"" + mapname + "\">");
     }

     /**
      * EndMap - end an image map specification
      */

     public void EndMap() {
        toClient.println("</MAP>");
     }

     /**
      * MapArea - define a hypertext area of an image map
      *
      * @param shape type of area
      * @param coords coordinates for the area
      * @param link the URL link for this area
      */

     public void MapArea(String shape, String coords, String link) {
        toClient.println("<AREA SHAPE=\"" + shape + "\" COORDS=\"" +
                           coords + "\" HREF=\"" + link + "\">");
     }

     // **********************************************************************
     // constructs that handle HTML text attributes and layout options
     // **********************************************************************

     /**
      * HeadingOne - a large section heading
      *
      * @param head string of text to use as heading
      *
      */

     public void HeadingOne(String head)
     {
        toClient.println("<H1> " + head +  " </H1>");
     }

     /**
      * HeadingTwo - a slightly less bold section heading
      *
      * @param head string of text to use as heading
      *
      */

     public void HeadingTwo(String head)
     {
        toClient.println("<H2> " + head +  " </H2>");
     }

     /**
      * HeadingThree - a mid-sized section heading
      *
      * @param head string of text to use as heading
      *
      */

     public void HeadingThree(String head)
     {
        toClient.println("<H3> " + head +  " </H3>");
     }

     /**
      * LineBreak - insert a linebreak between elements in the page
      *
      * @param none
      *
      */

     public void LineBreak()
     {
        toClient.println("<BR>");
     }

     /**
      * Center - start a block of centered information
      *
      * @param none
      */

      public void Center()
      {
         toClient.println("<CENTER>");
      }

     /**
      * EndCenter - end a block of centered information
      *
      * @param none
      */

     public void EndCenter()
     {
        toClient.println("</CENTER>");
     }

     /**
      * Paragraph - insert a paragraph break between elements in the page
      *
      * @param none
      *
      */

     public void Paragraph()
     {
        toClient.println("<P>");
     }

     /**
      * Paragraph - insert a paragraph of text in the page
      *
      * @param ptext the text of the paragraph
      *
      */

     public void Paragraph(String ptext)
     {
        toClient.println("<P> " + ptext +  "<P>");
     }

     /**
      * PreFormatted - insert a body of preformatted text in the page
      *
      * @param ptext the text
      *
      */

     public void PreFormatted(String ptext)
     {
        toClient.println("<PRE> " + ptext +  "</PRE>");
     }

     /**
      * PreFormatted - insert an Exception message as preformatted
      * text.
      *
      * @param ptext the text
      *
      */

     public void PreFormatted(Exception exception)
     {
        toClient.println("<PRE>");
        exception.printStackTrace(toClient);
        toClient.println("</PRE>");
     }

     /**
      * HorizontalRule - create a horizontal separator on the page
      */

     public void HorizontalRule()
     {
        toClient.println("<HR>");
     }

     /**
      * HorizontalRule - create a horizontal separator on the page
      *
      * @param width width of the rule as percent of page
      * @param thick thickness of the rule
      *
      */

     public void HorizontalRule(int width, int thick)
     {
        toClient.println("<HR width=\"" + width + "\" size=\"" + thick + "\">");
     }

     /**
      * HorizontalRule - create a horizontal separator on the page
      *
      * @param width width of the rule as percent of page
      * @param thick thickness of the rule
      * @param align on of {right|center|left}
      */

     public void HorizontalRule(int width, int thick, String align)
     {
        toClient.println("<HR width=\"" + width + "\" size=\"" +
                           thick + "\" ALIGN=\"" + align + "\">");
     }

    /* THE END */
 }





